# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This module implements a simulator for generating synthetic time series data.
"""

from __future__ import annotations

import copy
from datetime import timedelta
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
from scipy.stats import norm  # @manual

from kats.consts import TimeSeriesData

# A TimedeltaLike object represents a time offset.
# E.g., length of 1 day can be represented by timedelta(days=1),or by
# a timestamp offset 86400, or by the offset alias "1D" defined in pandas
TimedeltaLike = Union[timedelta, float, str]


def moving_average(a: np.ndarray, n: int = 3) -> np.ndarray:
    """Moving average of a numpy array.

    Args:
        a: numpy array.
        n: number of points over which we take moving avg.

    Returns:
        Array of moving average
    """
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1 :] / n


class Simulator:
    """TimeSeriesData simulator, to generate synthetic timeseries data.

    The Simulator currently supports generating synthetic time series
    using the STL, ARIMA models and also adds level and trend changepoints.

    Attributes:
        n: length of the time series.
        freq: desired frequency (e.g. daily, weekly) of a time series.
        start: start date of the time series.
    """

    def __init__(
        self,
        n: int = 100,
        freq: str = "D",
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        start: Optional[Any] = None,
    ) -> None:
        self.n = n
        self.freq = freq
        self.start = start

        # create time
        # pyre-fixme[4]: Attribute must be annotated.
        self.time = pd.date_range(
            start=start,
            freq=freq,
            periods=n,
        )

        # create the simulated time series
        # pyre-fixme[4]: Attribute must be annotated.
        self.timeseries = np.zeros(self.n)

    def arima_sim(
        self,
        ar: List[float],
        ma: List[float],
        mu: float = 0,
        sigma: float = 1,
        burnin: int = 10,
        d: int = 0,
        t: int = 0,
    ) -> TimeSeriesData:
        """Simulate data from ARIMA model.

        Data generation includes two steps:
        (1). Simulate data from ARMA(p', q) model

        The configuration of ARMA(p', q) model is:
        X_t = alpha_1 * X_{t-1} + ... + alpha_p * X_{t-p'}
                + 1 * epsilon_t + theta_1 * epsilon_{t-1} + ... + theta_q * epsilon_{t-q}

        (2). Add drift d
        d is the order of differencing
        p = p' - d for ARIMA(p, d, q)

        Args:
            ar: [alpha_1, ..., alpha_p'], coefficients of AR parameters.
                   p = len(alpha)
            ma: [theta_1, ..., theta_q], coefficients of MA parameters.
                   q = len(theta)
            epsilon: error terms follows normal distribution(mu, sigma).
            mu: mean of normal distribution for epsilon.
            sigma: standard dev of normal distribution for epsilon.
            burnin: number of data that will be dropped because lack of lagged data in the beginning.
            d: number of unit roots for non-stationary data.
            t: linear trend constant.

        Returns:
            ts: TimeSeries generated.
        Examples:
        >>> sim = Simulator(n=100, freq="MS", start = pd.to_datetime("2011-01-01 00:00:00"))
        >>> np.random.seed(100)
        >>> ts = sim.arima_sim(ar=[0.1, 0.05], ma = [0.04, 0.1], d = 1)
        """

        # validate params
        if sigma < 0 or sigma == 0:
            raise ValueError("Standard deviation of normal distribution must be positive")

        # Step 1: simulate ARMA(p, q)
        # get the max order of p' and q
        p_max = len(ar)

        # add theta_0=1 to ma coefficients
        ma = np.append(1, ma)
        q_max = len(ma)

        # simulate epsilon
        pq_max = max(p_max, q_max)
        epsilon = np.random.normal(mu, sigma, (self.n + pq_max + burnin, 1))

        # simulated data
        x = np.zeros((self.n + pq_max + burnin, 1))

        # initialization
        x[0] = epsilon[0]

        for i in range(1, x.shape[0]):
            AR = np.dot(ar[0 : min(i, p_max)], np.flip(x[i - min(i, p_max) : i], 0))
            MA = np.dot(
                ma[0 : min(i + 1, q_max)],
                np.flip(epsilon[i - min(i, q_max - 1) : i + 1], 0),
            )
            x[i] = AR + MA + t

        # Step 2: add drift, unit roots
        if d != 0:
            ARMA = x[-self.n :]
            m = ARMA.shape[0]
            temp = np.zeros((m + 1, 1))

            for _ in range(d):
                for j in range(m):
                    temp[j + 1] = ARMA[j] + temp[j]
                ARMA = temp[1:]

            # replacing data
            x[-self.n :] = temp[1:]

        # generate time
        time = pd.date_range(
            start=self.start,
            freq=self.freq,
            periods=self.n,
        )

        ts = TimeSeriesData(time=time, value=pd.Series(x[-self.n :].reshape(-1), copy=False))
        return ts

    def add_trend(self, magnitude: float, trend_type: str = "linear", multiply: bool = False) -> Simulator:
        """Add a trend component to the target time series for STL-based simulator.

        trend_type -  shape of the trend. {"linear","sigmoid"}

        Args:
            magnitude: slope of the trend, float.
            trend_type: linear or sigmoid, string.
            multiply: True if the trend is multiplicative, otherwise additive.

        Returns:
            The timeseries generated.
        """

        def component_gen(timepoints: np.ndarray) -> float:
            if trend_type == "sigmoid" or trend_type == "S":
                return magnitude * self.sigmoid(timepoints - 0.5)
            else:  # 'linear' trend by default
                return magnitude * timepoints

        return self._add_component(component_gen, multiply)

    def add_noise(
        self,
        magnitude: float = 1.0,
        multiply: bool = False,
    ) -> Simulator:

        """Add noise to the generated time series for STL-based simulator.

        Noise type is normal - noise will be generated from iid normal distribution;
        may consider adding noise generated by ARMA in the future if there're use cases.

        Args:
            magnitude: float.
            multiply: True if the noise is multiplicative, otherwise additive.

        Returns:
            Generated timeseries.
        """

        def component_gen(timepoints: np.ndarray) -> float:
            return magnitude * np.random.randn(len(timepoints))

        return self._add_component(component_gen, multiply)

    def add_seasonality(
        self,
        magnitude: float = 0.0,
        period: TimedeltaLike = "1D",
        multiply: bool = False,
    ) -> Simulator:
        """Add a seasonality component to the time series for STL-based simulator.

        Args:
            magnitude: slope of the trend, float.
            period: period of seasonality, timedelta.
            multiply: True if the seasonality is multiplicative, otherwise additive.

        Returns:
            Generated timeseries.
        """

        # pyre-fixme[6]: For 1st param expected `str` but got `Union[float, str,
        #  timedelta]`.
        period = self._convert_period(period)

        def component_gen(timepoints: np.ndarray) -> float:
            return magnitude * np.sin(2 * np.pi * timepoints)

        return self._add_component(component_gen, multiply, time_scale=period)

    @staticmethod
    def sigmoid(x: float) -> float:
        return 1 / (1 + np.exp(-10 * x))

    @staticmethod
    def _convert_period(period: str) -> float:
        """
        Convert TimedeltaLike object to time offset in seconds
        """
        # pyre-fixme[16]: `Optional` has no attribute `nanos`.
        return to_offset(period).nanos / 1e9

    def _add_component(
        self,
        # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
        component_gen: Callable,
        multiply: bool,
        time_scale: Optional[float] = None,
    ) -> Simulator:
        """
        Add a new component to the target time series.
        The component is defined by component_gen.
        """
        timestamps = self.time.values.astype(np.float64) / 1e9
        if time_scale is None:
            time_scale = timestamps[-1] - timestamps[0] + np.finfo(float).eps
        timepoints = (timestamps - timestamps[0]) / time_scale
        component = component_gen(timepoints)

        if multiply:
            self.timeseries *= 1 + component
        else:
            self.timeseries += component

        return self

    def stl_sim(self) -> TimeSeriesData:
        """Simulate time series data with seasonality, trend, and noise.

        Args:
            None.

        Returns:
            Generated timeseries.

        Example usage:
        >>> sim = Simulator(n=100, freq="1D", start = pd.to_datetime("2011-01-01"))
        >>> sim.add_trend(magnitude=10)
        >>> sim.add_seasonality(5, period=timedelta(days=7))
        >>> sim.add_noise(magnitude=2)
        >>> sim_ts = sim.stl_sim()
        """

        ts = TimeSeriesData(time=self.time, value=pd.Series(self.timeseries, copy=False))
        return ts

    def _adjust_length(self, length: int) -> None:
        """
        if given length is not compatible with other parameters, adjust the length
        """

        self.n = length
        self.time = pd.date_range(
            start=self.start,
            freq=self.freq,
            periods=self.n,
        )

    def _get_level_shift_y_val(
        self,
        random_seed: int = 100,
        cp_arr: Optional[List[int]] = None,
        level_arr: Optional[List[float]] = None,
        noise: float = 30,
        seasonal_period: int = 7,
        seasonal_magnitude: float = 3.0,
        anomaly_arr: Optional[List[int]] = None,
        z_score_arr: Optional[List[float]] = None,
    ) -> pd.Series:
        """
        generates values of a time series with multiple level shifts
        """
        if cp_arr is None:
            cp_arr = [100, 200, 350]
        if level_arr is None:
            level_arr = [1.35, 1.05, 1.35, 1.2]
        if anomaly_arr is None:
            anomaly_arr = []
        if z_score_arr is None:
            z_score_arr = []

        np.random.seed(seed=random_seed)

        # if cp_arr is not sorted, sort it
        cp_arr = sorted(cp_arr)

        # length of trend array should be one larger than cp array
        # so that there is a trend corresponding to every segment
        if len(level_arr) - len(cp_arr) != 1:
            raise ValueError(
                f"""
                Length of level array should be one greater than
                cp array. But we got
                cp_arr: {len(cp_arr)},
                level_arr: {len(level_arr)}
                """
            )

        if len(cp_arr) > 0 and cp_arr[-1] >= self.n:
            raise ValueError(f"Last cp {cp_arr[-1]} is greater than length {self.n}")

        cp_arr.append(self.n)
        cp_arr.insert(0, 0)

        y_val = np.ones(self.n)

        for i in range(len(cp_arr) - 1):
            cp_begin = cp_arr[i]
            cp_end = cp_arr[i + 1]

            y_val[cp_begin:cp_end] = norm.rvs(loc=level_arr[i], scale=noise, size=(cp_end - cp_begin))

        # add anomalies
        if len(anomaly_arr) != len(z_score_arr):
            raise ValueError(
                f"""
                Length of anomaly array should be equal to z_score array. But we got
                anomaly_arr: {len(anomaly_arr)},
                z_score_arr: {len(z_score_arr)}
                """
            )
        for arr_idx, y_idx in enumerate(anomaly_arr):
            if y_idx < 0 or y_idx >= self.n:
                raise ValueError(f"Anamaly point {y_idx} is out of range")
            # find out the mean value of this segment
            seg_idx = 0
            while y_idx > cp_arr[seg_idx]:
                seg_idx += 1
            y_val[y_idx] = level_arr[seg_idx - 1] + z_score_arr[arr_idx] * noise

        # add seasonality
        y_val += seasonal_magnitude * np.sin((np.pi / seasonal_period) * np.arange(self.n))

        return pd.Series(y_val, copy=False)

    def level_shift_sim(
        self,
        random_seed: int = 100,
        cp_arr: Optional[List[int]] = None,
        level_arr: Optional[List[float]] = None,
        noise: float = 30.0,
        seasonal_period: int = 7,
        seasonal_magnitude: float = 3.0,
        anomaly_arr: Optional[List[int]] = None,
        z_score_arr: Optional[List[float]] = None,
    ) -> TimeSeriesData:
        """Produces a time series with level shifts.

        The positions of the level shifts are indicated by the beginning and
        end changepoints.
        the duration of the first change is [first_cp_begin, first_cp_end],
        the duration of the second change point is [second_cp_begin, self.n]

        Args:
            cp_arr: Array of changepoint locations.
            level_arr: Array containing levels for each segment. Since the
                number of segments is one more than the number of changepoints,
                hence, the level arr should be longer than the cp_arr by one.
            noise: std. dev of random Gaussian noise added.
            seasonal_period: periodicity of the time series.
            seasonal_magnitude: amplitude of the seasonality. Set this to 0, if
                you want a time series without seasonality.
            anomaly_arr: locations where we introduce an anomalous point.
            z_score_arr: same length as anomaly arr. This is the z-score of the
                anomaly introduced at the location indicated by the anomaly_arr.

        Returns:
            Generated timeseries.

        Example Usage:
        >>> sim2 = Simulator(n=450, start="2018-01-01")
        >>> ts2 = sim2.level_shift_sim(
                cp_arr=[100, 200],
                level_arr=[3, 20, 2],
                noise=3,
                seasonal_period=7,
                seasonal_magnitude=3,
                anomaly_arr = [50, 150, 250],
                z_score_arr = [10, -10, 20],
        )
        """

        if cp_arr is None:
            cp_arr = [100, 200, 350]
        if level_arr is None:
            level_arr = [1.35, 1.05, 1.35, 1.2]
        if anomaly_arr is None:
            anomaly_arr = []
        if z_score_arr is None:
            z_score_arr = []

        # we need atleast points till second cp
        if len(cp_arr) > 0 and self.n < np.max(cp_arr):
            self._adjust_length(np.max(cp_arr))

        ts = TimeSeriesData(
            time=self.time,
            value=self._get_level_shift_y_val(
                random_seed,
                cp_arr,
                level_arr,
                noise,
                seasonal_period,
                seasonal_magnitude,
                anomaly_arr,
                z_score_arr,
            ),
        )

        return ts

    def level_shift_multivariate_indep_sim(
        self,
        cp_arr: Optional[List[int]] = None,
        level_arr: Optional[List[float]] = None,
        noise: float = 30.0,
        seasonal_period: int = 7,
        seasonal_magnitude: float = 3.0,
        anomaly_arr: Optional[List[int]] = None,
        z_score_arr: Optional[List[float]] = None,
        dim: int = 3,
    ) -> TimeSeriesData:
        """Produces a multivariate time series with level shifts.

        The positions of the level shifts are indicated by the beginning and
        end changepoints.
        the duration of the first change is [first_cp_begin, first_cp_end],
        the duration of the second change point is [second_cp_begin, self.n]
        The number of dimensions are indicated by dim.

        Args:
            cp_arr: Array of changepoint locations.
            level_arr: Array containing levels for each segment. Since the
                number of segments is one more than the number of changepoints,
                hence, the level arr should be longer than the cp_arr by one.
            noise: std. dev of random Gaussian noise added.
            seasonal_period: periodicity of the time series.
            seasonal_magnitude: amplitude of the seasonality. Set this to 0, if
                you want a time series without seasonality.
            anomaly_arr: locations where we introduce an anomalous point.
            z_score_arr: same length as anomaly arr. This is the z-score of the
                anomaly introduced at the location indicated by the anomaly_arr.
            dim: number of dimensions of the timeseries.

        Returns:
            Generated timeseries.
        """

        if cp_arr is None:
            cp_arr = [100, 200, 350]
        if level_arr is None:
            level_arr = [1.35, 1.05, 1.35, 1.2]
        if anomaly_arr is None:
            anomaly_arr = []
        if z_score_arr is None:
            z_score_arr = []

        # we need atleast points till second cp
        if self.n < np.max(cp_arr):
            self._adjust_length(np.max(cp_arr))

        df_dict = {"time": self.time}

        for i in range(dim):
            yval = self._get_level_shift_y_val(
                100 * (i + 1),
                cp_arr,
                level_arr,
                noise,
                seasonal_period,
                seasonal_magnitude,
                anomaly_arr,
                z_score_arr,
            )
            df_dict[f"value{i+1}"] = yval

        ts_df = pd.DataFrame(df_dict, copy=False)

        return TimeSeriesData(ts_df)

    def trend_shift_sim(
        self,
        random_seed: int = 15,
        cp_arr: Optional[List[int]] = None,
        trend_arr: Optional[List[float]] = None,
        intercept: float = 100.0,
        noise: float = 30.0,
        seasonal_period: int = 7,
        seasonal_magnitude: float = 3.0,
        anomaly_arr: Optional[List[int]] = None,
        z_score_arr: Optional[List[int]] = None,
    ) -> TimeSeriesData:
        """Produces a time series with multiple trend shifts and seasonality.

        This can be used as synthetic data to test trend changepoints
        first_cp_begin is where the trend change begins, and continues till the
        end.

        Args:
            random_seed: Seed, to reproduce the same time series.
            cp_arr: Array of changepoint locations.
            trend_arr: Array containing trends for each segment. Since the
                number of segments is one more than the number of changepoints,
                hence, the trend arr should be longer than the cp_arr by one.
            noise: std. dev of random Gaussian noise added.
            seasonal_period: periodicity of the time series.
            seasonal_magnitude: amplitude of the seasonality. Set this to 0, if
                you want a time series without seasonality.
            anomaly_arr: locations where we introduce an anomalous point.
            z_score_arr: same length as anomaly arr. This is the z-score of the
                anomaly introduced at the location indicated by the anomaly_arr.

        Returns:
            Generated timeseries.

        Example usage:
        >>> sim2 = Simulator(n=450, start="2018-01-01")
        >>> ts2 = sim2.trend_shift_sim(
                cp_arr=[100, 200],
                trend_arr=[3, 20, 2],
                intercept=30,
                noise=30,
                seasonal_period=7,
                seasonal_magnitude=3,
                anomaly_arr = [50, 150, 250],
                z_score_arr = [10, -10, 20],
        )
        """

        # initializing the lists inside the function since
        # mutable lists as defaults is bad practice that linter flags
        if cp_arr is None:
            cp_arr = [100]
        if trend_arr is None:
            trend_arr = [3.0, 30.0]
        if anomaly_arr is None:
            anomaly_arr = []
        if z_score_arr is None:
            z_score_arr = []

        # if cp_arr is not sorted, sort it
        cp_arr = sorted(cp_arr)

        # length of trend array should be one larger than cp array
        # so that there is a trend corresponding to every segment
        if len(trend_arr) - len(cp_arr) != 1:
            raise ValueError(
                f"""
                Length of trend array should be one greater than
                cp array. But we got
                cp_arr: {len(cp_arr)},
                trend_arr: {len(trend_arr)}
                """
            )

        if len(cp_arr) > 0 and cp_arr[-1] >= self.n:
            raise ValueError(f"Last cp {cp_arr[-1]} is greater than length {self.n}")

        cp_arr.append(self.n)
        cp_arr.insert(0, 0)

        y_val = np.full(self.n, intercept, dtype=float)

        for i in range(len(cp_arr) - 1):
            cp_begin = cp_arr[i]
            cp_end = cp_arr[i + 1]

            y_val[cp_begin:cp_end] = y_val[cp_begin:cp_end] + trend_arr[i] * np.arange(cp_begin, cp_end)

            if i > 0:
                delta_val = y_val[cp_begin] - y_val[cp_begin - 1]
                y_val[cp_begin:cp_end] -= delta_val

        # add seasonality
        y_val += seasonal_magnitude * np.sin((np.pi / seasonal_period) * np.arange(self.n))

        # add noise and anomalies
        noise_arr = norm.rvs(loc=0, scale=noise, size=self.n)
        if len(anomaly_arr) != len(z_score_arr):
            raise ValueError(
                f"""
                Length of anomaly array should be equal to z_score array. But we got
                anomaly_arr: {len(anomaly_arr)},
                z_score_arr: {len(z_score_arr)}
                """
            )
        for arr_idx, y_idx in enumerate(anomaly_arr):
            if y_idx < 0 or y_idx >= self.n:
                raise ValueError(f"Anomaly point {y_idx} is out of range")
            # pyre-fixme[16]: `Sequence` has no attribute `__setitem__`.
            noise_arr[y_idx] = z_score_arr[arr_idx] * noise

        y_val += noise_arr

        ts = TimeSeriesData(time=self.time, value=pd.Series(y_val, copy=False))

        return ts

    def _check_arr_inputs(
        self,
        ts: TimeSeriesData,
        loc_arr: Optional[List[int]],
        param_arr: Optional[List[float]],
        loc_param_diff: int = 1,
    ) -> None:
        """Private method, to check whether arguments to injected anomaly functions are valid.

        Args:
            ts: Given time series.
            loc_arr: Locations of anomaly/change point.
            param_arr: Parameters for each segment.
            loc_param_diff: Expected difference in length of loc and param arrays.
        """

        if loc_arr is None or len(loc_arr) == 0:
            raise ValueError("Changepoint Array cannot be empty")

        if param_arr is None or len(param_arr) == 0:
            raise ValueError("Parameter Array cannot be empty")

        len_ts = len(ts)

        # check if cp_arr is sorted
        if sorted(loc_arr) != loc_arr:
            raise ValueError("cp_arr must be sorted")

        # check if last cp is less than length of ts
        if loc_arr[-1] >= len_ts:
            raise ValueError(f"Last CP is at {loc_arr[-1]}, larger than length of ts, {len_ts}")

        if (len(loc_arr) - len(param_arr)) != loc_param_diff:
            raise ValueError(
                f"""
                Parameter array should be {loc_param_diff} longer than cp_arr, but
                cp_arr is {len(loc_arr)} long and parameter array is
                {len(param_arr)} long
                """
            )

    def inject_level_shift(
        self,
        ts_input: TimeSeriesData,
        cp_arr: Optional[List[int]] = None,
        level_arr: Optional[List[float]] = None,
    ) -> TimeSeriesData:
        """Produces Time Series with injected level shifts

        Args:
            ts_input: Input time series data.
            cp_arr: Array of changepoint locations.
            level_arr: Array containing levels for each segment. Since the
                number of segments is one more than the number of changepoints,
                hence, the level arr should be longer than the cp_arr by one.

        Returns:
            Timeseries with injected anomalies
        """

        if cp_arr is None or level_arr is None:
            return ts_input

        ts = copy.deepcopy(ts_input)
        self._check_arr_inputs(ts=ts, loc_arr=cp_arr, param_arr=level_arr)

        # check if two array sizes match, cp_arr is sorted, and less than length
        for i in range(len(cp_arr) - 1):
            start = cp_arr[i]
            end = cp_arr[i + 1]
            ts.value.values[start:end] += level_arr[i]

        return ts

    def inject_trend_shift(
        self,
        ts_input: TimeSeriesData,
        cp_arr: Optional[List[int]] = None,
        trend_arr: Optional[List[float]] = None,
    ) -> TimeSeriesData:
        """Produces Time Series with injected level shifts

        Args:
            ts_input: Input time series data.
            cp_arr: Array of changepoint locations.
            trend_arr: Array containing trends for each segment. Since the
                number of segments is one more than the number of changepoints,
                hence, the trend arr should be longer than the cp_arr by one.

        Returns:
            Timeseries with injected anomalies
        """

        if cp_arr is None or trend_arr is None:
            return ts_input

        ts = copy.deepcopy(ts_input)
        self._check_arr_inputs(ts=ts, loc_arr=cp_arr, param_arr=trend_arr)

        for i in range(len(cp_arr) - 1):
            start = cp_arr[i]
            end = cp_arr[i + 1]
            add_arr = trend_arr[i] * np.arange(end - start)
            ts.value.values[start:end] += add_arr

            # pyre-ignore Undefined attribute [16]: `float` has no attribute `__getitem__
            ts.value.values[end:] += add_arr[-1]

        return ts

    def inject_spikes(
        self,
        ts_input: TimeSeriesData,
        anomaly_arr: Optional[List[int]] = None,
        z_score_arr: Optional[List[float]] = None,
        epsilon_std_dev: float = 0,
    ) -> TimeSeriesData:
        """Produces time series with injected spikes.

        Args:
            ts_input: Input time series data.
            anomaly_arr: locations where we introduce an anomalous point.
            z_score_arr: same length as anomaly arr. This is the z-score of the
                anomaly introduced at the location indicated by the anomaly_arr.
            epsilon_std_dev: A small value of standard deviation, to produce anomalies
                even when the variance of the original time series is close to zero.
                This is useful for simulating anomalies in time series of nearly constant
                data.

        Returns:
            Timeseries with injected anomalies
        """

        if anomaly_arr is None or z_score_arr is None:
            return ts_input
        ts = copy.deepcopy(ts_input)
        self._check_arr_inputs(ts=ts, loc_arr=anomaly_arr, param_arr=z_score_arr, loc_param_diff=0)

        # get the std. dev of the time series
        ts_val = ts.value.values
        mov_avg = moving_average(ts_val, 3)
        std_dev = np.std(ts_val[2:] - mov_avg)

        for a_index, z_score in zip(anomaly_arr, z_score_arr):
            ts.value.values[a_index] += z_score * (std_dev + epsilon_std_dev)

        return ts



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import io
import os
import pkgutil
from datetime import datetime
from typing import List
from typing import cast
from unittest import TestCase

import numpy as np
import pandas as pd
import pytz
from dateutil import parser
from dateutil.relativedelta import relativedelta

from kats.compat.pandas import assert_frame_equal
from kats.compat.pandas import assert_index_equal
from kats.compat.pandas import assert_series_equal
from kats.consts import DEFAULT_TIME_NAME
from kats.consts import DEFAULT_VALUE_NAME
from kats.consts import TimeSeriesData
from kats.consts import TSIterator


def load_data(file_name: str) -> pd.DataFrame:
    ROOT = "kats"
    if "kats" in os.getcwd().lower():
        path = "data/"
    else:
        path = "kats/data/"
    data_object = pkgutil.get_data(ROOT, path + file_name)
    # pyre-fixme[6]: For 1st param expected `bytes` but got `Optional[bytes]`.
    return pd.read_csv(io.BytesIO(data_object), encoding="utf8")


TIME_COL_NAME = "ds"
VALUE_COL_NAME = "y"
MULTIVAR_VALUE_DF_COLS: List[str] = [VALUE_COL_NAME, VALUE_COL_NAME + "_1"]

EMPTY_DF = pd.DataFrame()
EMPTY_TIME_SERIES = pd.Series([], name=DEFAULT_TIME_NAME, dtype=float)
EMPTY_VALUE_SERIES = pd.Series([], name=DEFAULT_VALUE_NAME, dtype=float)
EMPTY_VALUE_SERIES_NO_NAME = pd.Series([], dtype=float)
EMPTY_TIME_DATETIME_INDEX = pd.DatetimeIndex(pd.Series([], dtype=object))
EMPTY_DF_WITH_COLS: pd.DataFrame = pd.concat([EMPTY_TIME_SERIES, EMPTY_VALUE_SERIES], axis=1)
NUM_YEARS_OFFSET = 12


class TimeSeriesBaseTest(TestCase):
    def setUp(self) -> None:
        # load Dataframes for testing
        self.AIR_DF = load_data("air_passengers.csv")
        self.AIR_DF_DATETIME = self.AIR_DF.copy(deep=True)
        self.AIR_DF_DATETIME.ds = self.AIR_DF_DATETIME.ds.apply(lambda x: parser.parse(x))
        self.AIR_DF_UNIXTIME = self.AIR_DF.copy(deep=True)
        self.AIR_DF_UNIXTIME.ds = self.AIR_DF_DATETIME.ds.apply(lambda x: (x - datetime(1970, 1, 1)).total_seconds())
        self.AIR_DF_WITH_DEFAULT_NAMES = self.AIR_DF.copy(deep=True)
        self.AIR_DF_WITH_DEFAULT_NAMES.columns = [DEFAULT_TIME_NAME, DEFAULT_VALUE_NAME]
        self.MULTIVAR_AIR_DF = self.AIR_DF.copy(deep=True)
        self.MULTIVAR_AIR_DF[VALUE_COL_NAME + "_1"] = self.MULTIVAR_AIR_DF.y * 2
        self.MULTIVAR_AIR_DF_DATETIME = self.MULTIVAR_AIR_DF.copy(deep=True)
        self.MULTIVAR_AIR_DF_DATETIME.ds = self.MULTIVAR_AIR_DF_DATETIME.ds.apply(lambda x: parser.parse(x))
        self.MULTIVAR_VALUE_DF = self.MULTIVAR_AIR_DF[MULTIVAR_VALUE_DF_COLS]
        self.AIR_TIME_SERIES = self.AIR_DF.ds
        self.AIR_TIME_SERIES_PD_DATETIME = pd.to_datetime(self.AIR_TIME_SERIES)
        self.AIR_TIME_SERIES_UNIXTIME = self.AIR_TIME_SERIES_PD_DATETIME.apply(
            lambda x: (x - datetime(1970, 1, 1)).total_seconds()
        )
        self.AIR_VALUE_SERIES = self.AIR_DF[VALUE_COL_NAME]
        self.AIR_TIME_DATETIME_INDEX = pd.DatetimeIndex(self.AIR_TIME_SERIES)


class TimeSeriesDataInitTest(TimeSeriesBaseTest):
    def setUp(self) -> None:
        super(TimeSeriesDataInitTest, self).setUp()
        # Univariate TimeSeriesData initialized from a pd.DataFrame
        self.ts_from_df = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        # Univariate TimeSeriesData initialized from a pd.DataFrame with time
        # as a datetime.datetime object
        self.ts_from_df_datetime = TimeSeriesData(df=self.AIR_DF_DATETIME, time_col_name=TIME_COL_NAME)
        # Univariate TimeSeriesData initialized from a pd.DataFrame with time
        # as unix time
        self.ts_from_df_with_unix = TimeSeriesData(
            df=self.AIR_DF_UNIXTIME,
            use_unix_time=True,
            unix_time_units="s",
            time_col_name=TIME_COL_NAME,
        )
        # Multivariate TimeSeriesData initialized from a pd.DataFrame
        self.ts_from_df_multi = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        # Multivariate TimeSeriesData initialized from a pd.DataFrame with time
        # as a datetime.datetime object
        self.ts_from_df_multi_datetime = TimeSeriesData(df=self.MULTIVAR_AIR_DF_DATETIME, time_col_name=TIME_COL_NAME)
        # Univariate TimeSeriesData initialized from two pd.Series with time
        # as a string
        self.ts_from_series_univar_no_datetime = TimeSeriesData(time=self.AIR_TIME_SERIES, value=self.AIR_VALUE_SERIES)
        # Univariate TimeSeriesData initialized from two pd.Series with time
        # as a pd.Timestamp
        self.ts_from_series_univar_with_datetime = TimeSeriesData(
            time=self.AIR_TIME_SERIES_PD_DATETIME, value=self.AIR_VALUE_SERIES
        )
        # Univariate TimeSeriesData initialized from two pd.Series with time
        # as unix time
        self.ts_from_series_with_unix = TimeSeriesData(
            time=self.AIR_TIME_SERIES_UNIXTIME,
            value=self.AIR_VALUE_SERIES,
            use_unix_time=True,
            unix_time_units="s",
            time_col_name=TIME_COL_NAME,
        )
        # Univariate TimeSeriesData initialized with time as a pd.Series and
        # value as a pd.DataFrame
        self.ts_from_series_and_df_univar = TimeSeriesData(
            time=self.AIR_TIME_SERIES, value=self.AIR_VALUE_SERIES.to_frame()
        )
        # Multivariate TimeSeriesData initialized from a pd.Series for time
        # and DataFrame for value
        self.ts_from_series_and_df_multivar = TimeSeriesData(time=self.AIR_TIME_SERIES, value=self.MULTIVAR_VALUE_DF)
        # Univariate TimeSeriesData initialized with time as a pd.DateTimeIndex
        # and value as a pd.Series
        self.ts_from_index_and_series_univar = TimeSeriesData(
            time=self.AIR_TIME_DATETIME_INDEX,
            value=self.AIR_VALUE_SERIES,
            time_col_name=TIME_COL_NAME,
        )
        # Multivariate TimeSeriesData initialized with time as a
        # pd.DateTimeIndex and value as a pd.DataFrame
        self.ts_from_index_and_series_multivar = TimeSeriesData(
            time=self.AIR_TIME_DATETIME_INDEX,
            value=self.MULTIVAR_VALUE_DF,
            time_col_name=TIME_COL_NAME,
        )
        # TimeSeriesData initialized from None Objects
        self.ts_df_none = TimeSeriesData(df=None)
        self.ts_time_none_and_value_none = TimeSeriesData(time=None, value=None)
        # TimeSeriesData initialized from Empty Objects
        self.ts_df_empty = TimeSeriesData(df=EMPTY_DF)
        self.ts_time_empty_value_empty = TimeSeriesData(time=EMPTY_TIME_SERIES, value=EMPTY_VALUE_SERIES)
        self.ts_time_empty_value_empty_no_name = TimeSeriesData(
            time=EMPTY_TIME_SERIES, value=EMPTY_VALUE_SERIES_NO_NAME
        )
        self.ts_time_empty_value_empty_df = TimeSeriesData(time=EMPTY_TIME_SERIES, value=EMPTY_DF)
        self.ts_time_empty_value_empty_df_with_cols = TimeSeriesData(time=EMPTY_TIME_SERIES, value=EMPTY_DF_WITH_COLS)

        # univariate data with missing time
        self.ts_univariate_missing = TimeSeriesData(
            df=pd.DataFrame(
                {
                    "time": ["2010-01-01", "2010-01-02", "2010-01-03", "2010-01-05"],
                    "value": [1, 2, 3, 4],
                }
            )
        )

        # multivariate data with missing time
        self.ts_multi_missing = TimeSeriesData(
            df=pd.DataFrame(
                {
                    "time": ["2010-01-01", "2010-01-02", "2010-01-03", "2010-01-05"],
                    "value1": [1, 2, 3, 4],
                    "value2": [4, 3, 2, 1],
                }
            )
        )

        # univariate data with unixtime in US/Pacific with time zone
        self.unix_list = (
            (pd.date_range("2020-03-01", "2020-03-10", tz="US/Pacific", freq="1d").astype(int) / 1e9)
            .astype(int)
            .to_list()
        )
        self.ts_univar_PST_tz = TimeSeriesData(
            df=pd.DataFrame({"time": self.unix_list, "value": [0] * 10}),
            use_unix_time=True,
            unix_time_units="s",
            tz="US/Pacific",
        )
        # univariate data with unixtime in US/Pacific without time zone
        self.ts_univar_PST = TimeSeriesData(
            df=pd.DataFrame({"time": self.unix_list, "value": [0] * 10}),
            use_unix_time=True,
            unix_time_units="s",
        )
        # univariate data with date str with tz
        date = ["2020-10-31", "2020-11-01", "2020-11-02"]
        self.ts_univar_str_date_tz = TimeSeriesData(
            df=pd.DataFrame({"time": date, "value": [0] * 3}),
            date_format="%Y-%m-%d",
            tz="US/Pacific",
        )
        # univariate data with date str without tz
        self.ts_univar_str_date = TimeSeriesData(
            df=pd.DataFrame({"time": date, "value": [0] * 3}),
            date_format="%Y-%m-%d",
        )

        # univariate data in US/Pacific Time Zone with missing data
        self.ts_univar_PST_missing_tz = TimeSeriesData(
            df=pd.DataFrame({"time": (self.unix_list[0:4] + self.unix_list[7:10]), "value": [0] * 7}),
            use_unix_time=True,
            unix_time_units="s",
            tz="US/Pacific",
        )

    # Testing univariate time series intialized from a DataFrame
    def test_init_from_df_univar(self) -> None:
        # DataFrame with string time
        assert_series_equal(self.ts_from_df.time, self.AIR_TIME_SERIES_PD_DATETIME)
        assert_series_equal(cast(pd.Series, self.ts_from_df.value), self.AIR_VALUE_SERIES)
        # DataFrame with datetime time
        assert_series_equal(self.ts_from_df_datetime.time, self.AIR_TIME_SERIES_PD_DATETIME)
        assert_series_equal(cast(pd.Series, self.ts_from_df_datetime.value), self.AIR_VALUE_SERIES)
        # DataFrame with unix time
        assert_series_equal(self.ts_from_df_with_unix.time, self.AIR_TIME_SERIES_PD_DATETIME)
        assert_series_equal(cast(pd.Series, self.ts_from_df_with_unix.value), self.AIR_VALUE_SERIES)

    # Testing multivariate time series initialized from a DataFrame
    def test_init_from_df_multi(self) -> None:
        assert_series_equal(self.ts_from_df_multi.time, self.AIR_TIME_SERIES_PD_DATETIME)
        assert_frame_equal(cast(pd.DataFrame, self.ts_from_df_multi.value), self.MULTIVAR_VALUE_DF)

    # Testing univariate time series initialized from a Series and Series/DataFrame
    def test_init_from_series_univar(self) -> None:
        # time and value from Series, with time as string
        assert_series_equal(
            self.ts_from_series_univar_no_datetime.time,
            self.AIR_TIME_SERIES_PD_DATETIME,
        )
        # time and value from Series, with time as pd.Timestamp
        assert_series_equal(
            self.ts_from_series_univar_with_datetime.time,
            self.AIR_TIME_SERIES_PD_DATETIME,
        )
        assert_series_equal(
            cast(pd.Series, self.ts_from_series_univar_no_datetime.value),
            self.AIR_VALUE_SERIES,
        )
        # time and value from Series, with time out of order and `sort_by_time=True`
        unsorted_df = self.AIR_DF.sample(frac=1)
        resorted_ts = TimeSeriesData(
            time=unsorted_df.ds,
            value=unsorted_df.y,
            time_col_name=TIME_COL_NAME,
            sort_by_time=True,
        )
        self.assertEqual(resorted_ts, self.ts_from_df)
        # time and value from Series, with time as unix time
        assert_series_equal(self.ts_from_series_with_unix.time, self.AIR_TIME_SERIES_PD_DATETIME)
        assert_series_equal(cast(pd.Series, self.ts_from_series_with_unix.value), self.AIR_VALUE_SERIES)
        # time from Series and value from DataFrame
        assert_series_equal(self.ts_from_series_and_df_univar.time, self.AIR_TIME_SERIES_PD_DATETIME)
        print(type(self.ts_from_series_and_df_univar.value))
        assert_series_equal(
            cast(pd.Series, self.ts_from_series_and_df_univar.value),
            self.AIR_VALUE_SERIES,
        )

    # Testing multivariate time series initialized from a Series/DataFrame
    def test_init_from_series_multivar(self) -> None:
        # Testing multivariate time series initialized from a
        assert_series_equal(self.ts_from_series_and_df_multivar.time, self.AIR_TIME_SERIES_PD_DATETIME)
        assert_frame_equal(
            cast(pd.DataFrame, self.ts_from_series_and_df_multivar.value),
            self.MULTIVAR_VALUE_DF,
        )

    # Testing univariate time series with time initialized as a
    # pd.DateTimeIndex
    def test_init_from_index_univar(self) -> None:
        assert_series_equal(self.ts_from_index_and_series_univar.time, self.AIR_TIME_SERIES_PD_DATETIME)
        assert_series_equal(
            cast(pd.Series, self.ts_from_index_and_series_univar.value),
            self.AIR_VALUE_SERIES,
        )

    # Testing multivariate time series with time initialized as a
    # pd.DateTimeIndex
    def test_init_from_index_multivar(self) -> None:
        assert_series_equal(
            self.ts_from_index_and_series_multivar.time,
            self.AIR_TIME_SERIES_PD_DATETIME,
        )
        assert_frame_equal(
            cast(pd.DataFrame, self.ts_from_index_and_series_multivar.value),
            self.MULTIVAR_VALUE_DF,
        )

    # Testing initialization from None Objects
    def test_none(self) -> None:
        # Testing initialization from None DataFrame
        assert_series_equal(self.ts_df_none.time, EMPTY_TIME_SERIES)
        assert_series_equal(cast(pd.Series, self.ts_df_none.value), EMPTY_VALUE_SERIES)
        # Testing initialization from two None Series
        assert_series_equal(self.ts_time_none_and_value_none.time, EMPTY_TIME_SERIES)
        assert_series_equal(cast(pd.Series, self.ts_time_none_and_value_none.value), EMPTY_VALUE_SERIES)

    # Testing initialization from Empty Objects
    def test_empty(self) -> None:
        # Testing intialization from empty DataFrame
        assert_series_equal(self.ts_df_empty.time, EMPTY_TIME_SERIES)
        assert_series_equal(cast(pd.Series, self.ts_df_empty.value), EMPTY_VALUE_SERIES)
        # Testing intialization from two empty Series
        assert_series_equal(self.ts_time_empty_value_empty.time, EMPTY_TIME_SERIES)
        assert_series_equal(cast(pd.Series, self.ts_time_empty_value_empty.value), EMPTY_VALUE_SERIES)
        # Testing intialization from two empty no name Series
        assert_series_equal(self.ts_time_empty_value_empty_no_name.time, EMPTY_TIME_SERIES)
        assert_series_equal(
            cast(pd.Series, self.ts_time_empty_value_empty_no_name.value),
            EMPTY_VALUE_SERIES,
        )

        # Make sure the time and value objects here have the default names
        self.assertEqual(self.ts_time_empty_value_empty_no_name.time.name, DEFAULT_TIME_NAME)
        self.assertEqual(self.ts_time_empty_value_empty_no_name.value.name, DEFAULT_VALUE_NAME)

        # Testing initialization from time as empty Series and value as empty
        # DataFrame
        assert_series_equal(self.ts_time_empty_value_empty_df.time, EMPTY_TIME_SERIES)
        assert_series_equal(cast(pd.Series, self.ts_time_empty_value_empty_df.value), EMPTY_VALUE_SERIES)
        # Testing initialization from time as empty Series and value as empty
        # DataFrame
        assert_series_equal(self.ts_time_empty_value_empty_df_with_cols.time, EMPTY_TIME_SERIES)
        assert_series_equal(
            cast(pd.Series, self.ts_time_empty_value_empty_df_with_cols.value),
            EMPTY_VALUE_SERIES,
        )

    # Testing incorrect initializations
    def test_incorrect_init_types(self) -> None:
        # Incorrect initialization with DF
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: Expected `Optional[pd.core.frame.DataFrame]` for 1st
            #  param but got `List[Variable[_T]]`.
            TimeSeriesData(df=[])

        # Incorrect initialization with value
        with self.assertRaises(ValueError):
            TimeSeriesData(time=self.AIR_TIME_SERIES, value=None)
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: Expected `Union[None, pd.core.frame.DataFrame,
            #  pd.core.series.Series]` for 2nd param but got `List[Variable[_T]]`.
            TimeSeriesData(time=self.AIR_TIME_SERIES, value=[])

        # Incorrect initialization with time
        with self.assertRaises(ValueError):
            TimeSeriesData(time=None, value=self.AIR_VALUE_SERIES)
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: Expected `Union[None,
            #  pd.core.indexes.datetimes.DatetimeIndex, pd.core.series.Series]` for 1st
            #  param but got `List[Variable[_T]]`.
            TimeSeriesData(time=[], value=self.AIR_VALUE_SERIES)

        # Incorrect initialization with time and value
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: Expected `Union[None,
            #  pd.core.indexes.datetimes.DatetimeIndex, pd.core.series.Series]` for 1st
            #  param but got `List[Variable[_T]]`.
            TimeSeriesData(time=[], value=[])

        # Incorrect initialization with value dtypes
        with self.assertRaises(ValueError):
            TimeSeriesData(time=self.AIR_TIME_SERIES, value=self.AIR_VALUE_SERIES.map(str))
        with self.assertRaises(ValueError):
            TimeSeriesData(time=self.AIR_TIME_SERIES, value=self.MULTIVAR_VALUE_DF.applymap(str))

    # Testing incorrect initializations
    def test_incorrect_init_lengths(self) -> None:
        # Incorrect initialization with different length time and values
        with self.assertRaises(ValueError):
            TimeSeriesData(time=self.AIR_TIME_SERIES, value=self.AIR_VALUE_SERIES[:-1])
        with self.assertRaises(ValueError):
            TimeSeriesData(time=self.AIR_TIME_SERIES[:-1], value=self.AIR_VALUE_SERIES)
        with self.assertRaises(ValueError):
            TimeSeriesData(time=self.AIR_TIME_SERIES, value=self.MULTIVAR_VALUE_DF[:-1])
        with self.assertRaises(ValueError):
            TimeSeriesData(time=self.AIR_TIME_SERIES[:-1], value=self.MULTIVAR_VALUE_DF)

    # Testing DataFrame conversion
    def test_to_dataframe(self) -> None:
        # Univariate case
        assert_frame_equal(self.ts_from_df.to_dataframe(), self.AIR_DF_DATETIME)
        # Multivariate case
        assert_frame_equal(self.ts_from_df_multi_datetime.to_dataframe(), self.MULTIVAR_AIR_DF_DATETIME)
        # Series Cases
        assert_frame_equal(self.ts_from_series_univar_no_datetime.to_dataframe(), self.AIR_DF_DATETIME)
        assert_frame_equal(
            self.ts_from_series_univar_with_datetime.to_dataframe(),
            self.AIR_DF_DATETIME,
        )
        # Series/DataFrame Cases
        assert_frame_equal(self.ts_from_series_and_df_univar.to_dataframe(), self.AIR_DF_DATETIME)
        assert_frame_equal(
            self.ts_from_series_and_df_multivar.to_dataframe(),
            self.MULTIVAR_AIR_DF_DATETIME,
        )
        # Empty/None Cases
        assert_frame_equal(self.ts_df_none.to_dataframe(), EMPTY_DF_WITH_COLS)
        assert_frame_equal(self.ts_time_none_and_value_none.to_dataframe(), EMPTY_DF_WITH_COLS)
        assert_frame_equal(self.ts_df_empty.to_dataframe(), EMPTY_DF_WITH_COLS)
        assert_frame_equal(self.ts_time_empty_value_empty.to_dataframe(), EMPTY_DF_WITH_COLS)
        assert_frame_equal(self.ts_time_empty_value_empty_df.to_dataframe(), EMPTY_DF_WITH_COLS)

    # Testing Data Interpolate
    def test_interpolate(self) -> None:
        # univariate
        self.assertEqual(
            self.ts_univariate_missing.interpolate(freq="D", method="linear"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value": [1, 2, 3, 3.5, 4],
                    }
                )
            ),
        )

        self.assertEqual(
            self.ts_univariate_missing.interpolate(freq="D", method="ffill"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value": [1, 2, 3, 3, 4],
                    }
                )
            ),
        )

        self.assertEqual(
            self.ts_univariate_missing.interpolate(freq="D", method="bfill"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value": [1, 2, 3, 4, 4],
                    }
                )
            ),
        )

        # check pd.interpolate method that is not explicitly defined in TimeSeriesData code
        self.assertEqual(
            self.ts_univariate_missing.interpolate(freq="D", method="nearest"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value": [1, 2, 3, 3, 4],
                    }
                )
            ),
        )

        # check pd.interpolate method that is not explicitly defined in TimeSeriesData code
        # and has additional parameters
        self.assertEqual(
            self.ts_univariate_missing.interpolate(freq="D", method="polynomial", order=3),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value": [1, 2, 3, 3.75, 4],
                    }
                )
            ),
        )

        # check pd.interpolate method that is not explicitly defined in TimeSeriesData code
        # and has additional parameters
        self.assertEqual(
            self.ts_univariate_missing.interpolate(freq="D", method="polynomial", order=1),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value": [1, 2, 3, 3.5, 4],
                    }
                )
            ),
        )

        # check method that is not in pd.interpolate and have not explicitly defined in TimeSeriesData code
        with self.assertRaises(ValueError):
            self.ts_univariate_missing.interpolate(freq="D", method="bad_input_should_fail")

        # multivariate
        self.assertEqual(
            self.ts_multi_missing.interpolate(freq="D", method="linear"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value1": [1, 2, 3, 3.5, 4],
                        "value2": [4, 3, 2, 1.5, 1],
                    }
                )
            ),
        )

        self.assertEqual(
            self.ts_multi_missing.interpolate(freq="D", method="ffill"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value1": [1, 2, 3, 3, 4],
                        "value2": [4, 3, 2, 2, 1],
                    }
                )
            ),
        )

        self.assertEqual(
            self.ts_multi_missing.interpolate(freq="D", method="bfill"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value1": [1, 2, 3, 4, 4],
                        "value2": [4, 3, 2, 1, 1],
                    }
                )
            ),
        )

        # test with no frequency given univariate
        self.assertEqual(
            self.ts_univariate_missing.interpolate(method="linear"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value": [1, 2, 3, 3.5, 4],
                    }
                )
            ),
        )

        # no frequency given, for multivariate
        self.assertEqual(
            self.ts_multi_missing.interpolate(method="linear"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value1": [1, 2, 3, 3.5, 4],
                        "value2": [4, 3, 2, 1.5, 1],
                    }
                )
            ),
        )

        # test methods that are not explicitly defined in TimeSeriesData
        self.assertEqual(
            self.ts_multi_missing.interpolate(method="time"),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value1": [1, 2, 3, 3.5, 4],
                        "value2": [4, 3, 2, 1.5, 1],
                    }
                )
            ),
        )

        # test methods that are not explicitly defined in TimeSeriesData
        # with additional arguments
        self.assertEqual(
            self.ts_multi_missing.interpolate(method="polynomial", order=3),
            TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": [
                            "2010-01-01",
                            "2010-01-02",
                            "2010-01-03",
                            "2010-01-04",
                            "2010-01-05",
                        ],
                        "value1": [1, 2, 3, 3.75, 4],
                        "value2": [4, 3, 2, 1.25, 1],
                    }
                )
            ),
        )

        # check method that is not in pd.interpolate and have not explicitly defined in TimeSeriesData code
        with self.assertRaises(ValueError):
            self.ts_multi_missing.interpolate(freq="D", method="bad_input_should_fail")

    # Testing Data interpolate with base
    def test_interpolate_base(self) -> None:
        # create time series with missing data
        np.random.seed(0)
        x = np.random.normal(0.5, 3, 998)
        time_val0 = list(pd.date_range(start="2018-02-03 14:59:59", freq="1800s", periods=1000))
        time_val = time_val0[:300] + time_val0[301:605] + time_val0[606:]
        ts0 = TimeSeriesData(pd.DataFrame({"time": time_val, "value": pd.Series(x)}))

        # calculate frequency first
        frequency = str(int(ts0.infer_freq_robust().total_seconds())) + "s"

        # without base value, interpolate won't work, will return all NaN
        # this is because start time is not from "**:00:00" or "**:30:00" type.
        self.assertEqual(
            ts0.interpolate(freq=frequency).to_dataframe().fillna(0).value.sum(),
            0,
        )
        # with base value, will start from "**:59:59" ("**:00:00" - 1 sec)
        # or "**:29:59" ("**:30:00" -1 sec).
        self.assertEqual(
            ts0.interpolate(freq=frequency, base=-1).to_dataframe().isna().value.sum(),
            0,
        )

        # second example, base = 4
        time_val0 = list(pd.date_range(start="2018-02-03 14:00:04", freq="1800s", periods=1000))
        time_val = time_val0[:300] + time_val0[301:605] + time_val0[606:]
        ts0 = TimeSeriesData(pd.DataFrame({"time": time_val, "value": pd.Series(x)}))

        frequency = str(int(ts0.infer_freq_robust().total_seconds())) + "s"

        self.assertEqual(
            ts0.interpolate(freq=frequency).to_dataframe().fillna(0).value.sum(),
            0,
        )
        self.assertEqual(
            ts0.interpolate(freq=frequency, base=4).to_dataframe().isna().value.sum(),
            0,
        )

    def test_to_array(self) -> None:
        # Univariate case
        np.testing.assert_array_equal(self.ts_from_df.to_array(), self.AIR_DF_DATETIME.to_numpy())
        # Multivariate case
        np.testing.assert_array_equal(
            self.ts_from_df_multi_datetime.to_array(),
            self.MULTIVAR_AIR_DF_DATETIME.to_numpy(),
        )
        # Series Cases
        np.testing.assert_array_equal(
            self.ts_from_series_univar_no_datetime.to_array(),
            self.AIR_DF_DATETIME.to_numpy(),
        )
        np.testing.assert_array_equal(
            self.ts_from_series_univar_with_datetime.to_array(),
            self.AIR_DF_DATETIME.to_numpy(),
        )
        # Series/DataFrame Cases
        np.testing.assert_array_equal(
            self.ts_from_series_and_df_univar.to_array(),
            self.AIR_DF_DATETIME.to_numpy(),
        )
        np.testing.assert_array_equal(
            self.ts_from_series_and_df_multivar.to_array(),
            self.MULTIVAR_AIR_DF_DATETIME.to_numpy(),
        )
        # Empty/None Cases
        np.testing.assert_array_equal(self.ts_df_none.to_array(), np.empty)
        np.testing.assert_array_equal(self.ts_time_none_and_value_none.to_array(), np.empty)
        np.testing.assert_array_equal(self.ts_df_empty.to_array(), np.empty)
        np.testing.assert_array_equal(self.ts_time_empty_value_empty.to_array(), np.empty)
        np.testing.assert_array_equal(self.ts_time_empty_value_empty_df.to_array(), np.empty)

    def test_tz(self) -> None:
        self.ts_univar_PST_tz.validate_data(validate_frequency=True, validate_dimension=True)
        self.assertEqual(self.ts_univar_PST_tz.freq_to_timedelta(), pd.Timedelta("1d"))
        self.assertEqual(self.ts_univar_PST_tz.tz(), pytz.timezone("US/Pacific"))
        self.assertTrue(
            (np.array(self.unix_list) == (self.ts_univar_PST_tz.time.values.astype(int) / 1e9).astype(int)).all()
        )

        with self.assertRaisesRegex(ValueError, "Only constant frequency is supported for time!"):
            self.ts_univar_PST.validate_data(validate_frequency=True, validate_dimension=True)

        self.ts_univar_str_date.validate_data(validate_frequency=True, validate_dimension=True)
        self.assertEqual(self.ts_univar_str_date.freq_to_timedelta(), pd.Timedelta("1d"))

        self.ts_univar_str_date_tz.validate_data(validate_frequency=True, validate_dimension=True)
        self.assertEqual(self.ts_univar_str_date_tz.freq_to_timedelta(), pd.Timedelta("1d"))
        self.assertEqual(self.ts_univar_PST_tz.tz(), pytz.timezone("US/Pacific"))

        # test ambiguous
        tsd = TimeSeriesData(
            df=pd.DataFrame(
                {
                    "time": [
                        "2018-10-28 01:30:00",
                        "2018-10-28 02:00:00",
                        "2018-10-28 02:30:00",
                        "2018-10-28 02:00:00",
                        "2018-10-28 02:30:00",
                        "2018-10-28 03:00:00",
                        "2018-10-28 03:30:00",
                    ],
                    "value": [0] * 7,
                }
            ),
            tz="CET",
            tz_ambiguous="infer",
        )
        tsd.validate_data(validate_frequency=True, validate_dimension=True)

        # test nonexistent
        tsd = TimeSeriesData(
            df=pd.DataFrame(
                {
                    "time": [
                        "2020-03-08 02:00:00",
                        "2020-03-08 02:30:00",
                        "2020-03-08 03:00:00",
                    ],
                    "value": [0] * 3,
                }
            ),
            tz="US/Pacific",
            tz_nonexistent="shift_forward",
        )

    def test_infer_freq_robust(self) -> None:
        self.assertEqual(
            self.ts_univariate_missing.infer_freq_robust(),
            pd.Timedelta(value=1, unit="D"),
        )

        self.assertEqual(
            self.ts_univar_PST_missing_tz.infer_freq_robust(),
            pd.Timedelta(value=1, unit="D"),
        )

    def test_is_data_missing(self) -> None:
        self.assertEqual(self.ts_univariate_missing.is_data_missing(), True)

        self.assertEqual(self.ts_univar_PST_missing_tz.is_data_missing(), True)

        self.assertEqual(self.ts_from_series_and_df_univar.is_data_missing(), False)

        self.assertEqual(self.ts_from_series_and_df_multivar.is_data_missing(), False)

    def test_min_max_values(self) -> None:
        # test min/max value for univariate
        self.assertEqual(self.ts_from_df.min, np.nanmin(self.ts_from_df.value.values))
        self.assertEqual(self.ts_from_df.max, np.nanmax(self.ts_from_df.value.values))

        # test min/max value for multivariate
        self.assertEqual(
            # pyre-fixme[16]: `float` has no attribute `equals`.
            self.ts_from_df_multi.min.equals(self.ts_from_df_multi.value.min(skipna=True)),
            True,
        )
        self.assertEqual(
            # pyre-fixme[16]: Item `float` of `Union[float, Series]` has no
            #  attribute `equals`.
            self.ts_from_df_multi.max.equals(self.ts_from_df_multi.value.max(skipna=True)),
            True,
        )

        # test min/max value for empty TS
        empty_ts = TimeSeriesData(pd.DataFrame())
        self.assertEqual(np.isnan(empty_ts.min), True)
        self.assertEqual(np.isnan(empty_ts.max), True)

        # test if min/max changes if values are re-assigned for univariate
        ts_from_df_new = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        new_val = np.random.randn(len(self.AIR_DF))
        ts_from_df_new.value = pd.Series(new_val)
        self.assertEqual(ts_from_df_new.min, np.min(new_val))
        self.assertEqual(ts_from_df_new.max, np.max(new_val))

        # test if min/max changes if values are re-assigned with NaNs for univariate
        new_val[-1] = np.nan
        ts_from_df_new.value = pd.Series(new_val)
        self.assertEqual(ts_from_df_new.min, np.nanmin(new_val))
        self.assertEqual(ts_from_df_new.max, np.nanmax(new_val))

        # test min/max changes if values are re-assigned for multivariate
        ts_from_df_multi_new = TimeSeriesData(self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        new_val_multi = np.random.randn(self.MULTIVAR_VALUE_DF.shape[0], self.MULTIVAR_VALUE_DF.shape[1] - 1)
        ts_from_df_multi_new.value = pd.DataFrame(new_val_multi)
        self.assertEqual(
            # pyre-fixme[16]: Item `float` of `Union[float, Series]` has no
            #  attribute `equals`.
            ts_from_df_multi_new.min.equals(pd.DataFrame(new_val_multi).min()),
            True,
        )
        self.assertEqual(
            # pyre-fixme[16]: Item `float` of `Union[float, Series]` has no
            #  attribute `equals`.
            ts_from_df_multi_new.max.equals(pd.DataFrame(new_val_multi).max()),
            True,
        )

        # test min/max changes if values are re-assigned with NaNs for multivariate
        new_val_multi[0] = np.nan
        ts_from_df_multi_new.value = pd.DataFrame(new_val_multi)
        self.assertEqual(
            # pyre-fixme[16]: Item `float` of `Union[float, Series]` has no
            #  attribute `equals`.
            ts_from_df_multi_new.min.equals(pd.DataFrame(new_val_multi).min(skipna=True)),
            True,
        )
        self.assertEqual(
            # pyre-fixme[16]: Item `float` of `Union[float, Series]` has no
            #  attribute `equals`.
            ts_from_df_multi_new.max.equals(pd.DataFrame(new_val_multi).max(skipna=True)),
            True,
        )


class TimeSeriesDataOpsTest(TimeSeriesBaseTest):
    def setUp(self) -> None:
        super(TimeSeriesDataOpsTest, self).setUp()
        # Creating DataFrames
        # DataFrame with date offset
        transformed_df_date = self.AIR_DF_DATETIME.copy(deep=True)
        transformed_df_date.ds = transformed_df_date.ds.apply(lambda x: x + relativedelta(years=NUM_YEARS_OFFSET))
        transformed_df_date_concat = self.AIR_DF.append(transformed_df_date, ignore_index=True)
        transformed_df_date_double = self.AIR_DF_DATETIME.copy(deep=True)
        transformed_df_date_double.ds = transformed_df_date.ds.apply(
            lambda x: x + relativedelta(years=NUM_YEARS_OFFSET * 2)
        )
        transformed_df_date_concat_double = self.AIR_DF.append(transformed_df_date_double, ignore_index=True)
        # DataFrames with value offset
        transformed_df_value = self.AIR_DF.copy(deep=True)
        transformed_df_value.y = transformed_df_value.y.apply(lambda x: x * 2)
        transformed_df_value_inv = self.AIR_DF.copy(deep=True)
        transformed_df_value_inv.y = transformed_df_value_inv.y.apply(lambda x: x * -1)
        # DataFrame with date and value offset
        transformed_df_date_and_value = transformed_df_date.copy(deep=True)
        transformed_df_date_and_value.y = transformed_df_date_and_value.y.apply(lambda x: x * 2)
        # DataFrame with date offset (multivariate)
        transformed_df_date_multi = transformed_df_date.copy(deep=True)
        transformed_df_date_multi[VALUE_COL_NAME + "_1"] = transformed_df_date_multi.y * 2
        transformed_df_date_concat_multi = self.MULTIVAR_AIR_DF.append(transformed_df_date_multi, ignore_index=True)
        transformed_df_date_concat_mixed = self.MULTIVAR_AIR_DF_DATETIME.append(transformed_df_date)
        transformed_df_date_double_multi = transformed_df_date_double.copy(deep=True)
        transformed_df_date_double_multi[VALUE_COL_NAME + "_1"] = transformed_df_date_double_multi.y * 2
        transformed_df_date_concat_double_multi = self.MULTIVAR_AIR_DF.append(
            transformed_df_date_double_multi, ignore_index=True
        )
        transformed_df_date_concat_double_mixed = self.MULTIVAR_AIR_DF_DATETIME.append(transformed_df_date_double)
        # DataFrame with value offset (multivariate)
        transformed_df_value_none_multi = self.MULTIVAR_AIR_DF.copy(deep=True)
        transformed_df_value_none_multi.y = transformed_df_value_none_multi.y_1
        transformed_df_value_none_multi.y_1 = np.nan
        # DataFrame with date and value offset (multivariate)
        transformed_df_date_and_value_multi = transformed_df_date_and_value.copy(deep=True)
        transformed_df_date_and_value_multi[VALUE_COL_NAME + "_1"] = transformed_df_date_and_value_multi.y * 2
        # DataFrame with all constant values
        df_zeros = self.AIR_DF.copy(deep=True)
        df_zeros.y.values[:] = 0
        df_ones = self.AIR_DF.copy(deep=True)
        df_ones.y.values[:] = 1
        df_twos = df_ones.copy(deep=True)
        df_twos.y.values[:] = 2
        df_neg_ones = self.AIR_DF.copy(deep=True)
        df_neg_ones.y.values[:] = -1
        df_ones_multi = df_ones.copy(deep=True)
        df_ones_multi[VALUE_COL_NAME + "_1"] = df_ones_multi.y * 2

        # Creating TimeSeriesData objects
        # Univariate TimeSeriesData initialized from a pd.DataFrame
        self.ts_univ_1 = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_univ_2 = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_univ_default_names = TimeSeriesData(df=self.AIR_DF_WITH_DEFAULT_NAMES)
        self.ts_univ_default_names_2 = TimeSeriesData(df=self.AIR_DF_WITH_DEFAULT_NAMES)

        # Multivariate TimeSeriesData initialized from a pd.DataFrame
        self.ts_multi_1 = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_multi_2 = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)

        # TimeSeriesData with date offset
        self.ts_date_transform_univ = TimeSeriesData(df=transformed_df_date, time_col_name=TIME_COL_NAME)
        self.ts_date_transform_concat_univ = TimeSeriesData(df=transformed_df_date_concat, time_col_name=TIME_COL_NAME)
        self.ts_date_transform_double_univ = TimeSeriesData(df=transformed_df_date_double, time_col_name=TIME_COL_NAME)
        self.ts_date_transform_concat_double_univ = TimeSeriesData(
            df=transformed_df_date_concat_double, time_col_name=TIME_COL_NAME
        )
        # TimeSeriesData with date offset (multivariate)
        self.ts_date_transform_multi = TimeSeriesData(df=transformed_df_date_multi, time_col_name=TIME_COL_NAME)
        self.ts_date_transform_concat_multi = TimeSeriesData(
            df=transformed_df_date_concat_multi, time_col_name=TIME_COL_NAME
        )
        self.ts_date_transform_concat_mixed = TimeSeriesData(
            df=transformed_df_date_concat_mixed, time_col_name=TIME_COL_NAME
        )
        self.ts_date_transform_double_multi = TimeSeriesData(
            df=transformed_df_date_double_multi, time_col_name=TIME_COL_NAME
        )
        self.ts_date_transform_concat_double_multi = TimeSeriesData(
            df=transformed_df_date_concat_double_multi, time_col_name=TIME_COL_NAME
        )
        self.ts_date_transform_concat_double_mixed = TimeSeriesData(
            df=transformed_df_date_concat_double_mixed, time_col_name=TIME_COL_NAME
        )
        # TimeSeriesData with value offset
        self.ts_value_transform_univ = TimeSeriesData(df=transformed_df_value, time_col_name=TIME_COL_NAME)
        self.ts_value_transform_inv_univ = TimeSeriesData(df=transformed_df_value_inv, time_col_name=TIME_COL_NAME)
        # TimeSeriesData with value offset (multivariate)
        self.ts_value_transform_none_multi = TimeSeriesData(
            df=transformed_df_value_none_multi, time_col_name=TIME_COL_NAME
        )
        # TimeSeriesData with date and value offset
        self.ts_date_and_value_transform_univ = TimeSeriesData(
            df=transformed_df_date_and_value, time_col_name=TIME_COL_NAME
        )
        # TimeSeriesData with date and value offset (multivariate)
        self.ts_date_and_value_transform_multi = TimeSeriesData(
            df=transformed_df_date_and_value_multi, time_col_name=TIME_COL_NAME
        )
        # TimeSeriesData object with all constant values
        self.ts_zero = TimeSeriesData(df=df_zeros, time_col_name=TIME_COL_NAME)
        self.ts_ones = TimeSeriesData(df=df_ones, time_col_name=TIME_COL_NAME)
        self.ts_twos = TimeSeriesData(df=df_twos, time_col_name=TIME_COL_NAME)
        self.ts_neg_ones = TimeSeriesData(df=df_neg_ones, time_col_name=TIME_COL_NAME)
        self.ts_ones_multi = TimeSeriesData(df=df_ones_multi, time_col_name=TIME_COL_NAME)
        # Empty TimeSeriesData Object
        self.ts_empty = TimeSeriesData(df=EMPTY_DF)
        self.ts_empty_with_cols = TimeSeriesData(df=EMPTY_DF_WITH_COLS, time_col_name=TIME_COL_NAME)
        # Copies for Extended objects
        self.ts_univ_extend = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_univ_extend_2 = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_univ_extend_err = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_multi_extend = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_multi_extend_2 = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_multi_extend_3 = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_multi_extend_4 = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_multi_extend_err = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_multi_extend_err_2 = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)
        self.ts_empty_extend = TimeSeriesData(df=EMPTY_DF)
        self.ts_empty_extend_err = TimeSeriesData(df=EMPTY_DF)

        # Other values
        self.length = len(self.AIR_DF)

    def test_eq(self) -> None:
        # Univariate equality
        self.assertTrue(self.ts_univ_1 == self.ts_univ_2)
        # Multivariate equality
        self.assertTrue(self.ts_multi_1 == self.ts_multi_2)
        # Univariate inequality
        self.assertFalse(self.ts_univ_1 == self.ts_date_transform_univ)
        self.assertFalse(self.ts_univ_1 == self.ts_value_transform_univ)
        self.assertFalse(self.ts_univ_1 == self.ts_date_and_value_transform_univ)
        # Multivariate inequality
        self.assertFalse(self.ts_multi_1 == self.ts_date_transform_multi)
        self.assertFalse(self.ts_multi_1 == self.ts_value_transform_none_multi)
        self.assertFalse(self.ts_multi_1 == self.ts_date_and_value_transform_multi)
        # Univariate vs. Multivariate inequality
        self.assertFalse(self.ts_univ_1 == self.ts_multi_1)
        self.assertFalse(self.ts_multi_1 == self.ts_univ_1)

    def test_ne(self) -> None:
        # Univariate equality
        self.assertFalse(self.ts_univ_1 != self.ts_univ_2)
        # Multivariate equality
        self.assertFalse(self.ts_multi_1 != self.ts_multi_2)
        # Univariate inequality
        self.assertTrue(self.ts_univ_1 != self.ts_date_transform_univ)
        self.assertTrue(self.ts_univ_1 != self.ts_value_transform_univ)
        self.assertTrue(self.ts_univ_1 != self.ts_date_and_value_transform_univ)
        # Multivariate inequality
        self.assertTrue(self.ts_multi_1 != self.ts_date_transform_multi)
        self.assertTrue(self.ts_multi_1 != self.ts_value_transform_none_multi)
        self.assertTrue(self.ts_multi_1 != self.ts_date_and_value_transform_multi)
        # Univariate vs. Multivariate inequality
        self.assertTrue(self.ts_univ_1 != self.ts_multi_1)
        self.assertTrue(self.ts_multi_1 != self.ts_univ_1)

    def test_add(self) -> None:
        # Add same DataFrames
        self.assertEqual(self.ts_univ_1 + self.ts_univ_2, self.ts_value_transform_univ)
        # Add different DataFrames
        self.assertEqual(self.ts_univ_1 + self.ts_value_transform_inv_univ, self.ts_zero)
        # Add Univariate and Multivariate DataFrames
        self.assertEqual(self.ts_univ_1 + self.ts_multi_1, self.ts_value_transform_none_multi)
        # Empty Case
        self.assertEqual(self.ts_empty + self.ts_empty, self.ts_empty)
        # Add DataFrames with different dates
        with self.assertRaises(ValueError):
            self.ts_univ_1 + self.ts_date_transform_univ

    def test_sub(self) -> None:
        # Subtract same DataFrames
        self.assertEqual(self.ts_univ_1 - self.ts_univ_2, self.ts_zero)
        # Subtract different DataFrames
        self.assertEqual(
            self.ts_univ_1 - self.ts_value_transform_inv_univ,
            self.ts_value_transform_univ,
        )
        # Subtract Univariate and Multivariate DataFrames
        self.assertEqual(
            self.ts_multi_1 - self.ts_value_transform_inv_univ,
            self.ts_value_transform_none_multi,
        )
        # Empty Case
        self.assertEqual(self.ts_empty - self.ts_empty, self.ts_empty)
        # Subtract DataFrames with different dates
        with self.assertRaises(ValueError):
            self.ts_univ_1 - self.ts_date_transform_univ

    def test_div(self) -> None:
        # Divide same DataFrames
        self.assertEqual(self.ts_univ_1 / self.ts_univ_2, self.ts_ones)
        # Divide different DataFrames
        self.assertEqual(self.ts_univ_1 / self.ts_value_transform_inv_univ, self.ts_neg_ones)
        # Divide Univariate and Multivariate DataFrames
        self.assertEqual(
            self.ts_value_transform_univ / self.ts_ones_multi,
            self.ts_value_transform_none_multi,
        )
        # Empty Case
        self.assertEqual(self.ts_empty / self.ts_empty, self.ts_empty)
        # Divide DataFrames with different dates
        with self.assertRaises(ValueError):
            self.ts_univ_1 / self.ts_date_transform_univ

    def test_mul(self) -> None:
        # Multiply same DataFrames
        self.assertEqual(self.ts_ones * self.ts_ones, self.ts_ones)
        # Multiply different DataFrames
        self.assertEqual(self.ts_univ_1 * self.ts_twos, self.ts_value_transform_univ)
        # Multiply Univariate and Multivariate DataFrames
        self.assertEqual(self.ts_multi_1 * self.ts_twos, self.ts_value_transform_none_multi)
        # Empty Case
        self.assertEqual(self.ts_empty * self.ts_empty, self.ts_empty)
        # Multiply DataFrames with different dates
        with self.assertRaises(ValueError):
            self.ts_univ_1 * self.ts_date_transform_univ

    def test_len(self) -> None:
        # Normal case
        self.assertEqual(len(self.ts_univ_1), self.length)
        # Empty case
        self.assertEqual(len(self.ts_empty), 0)

    def test_empty(self) -> None:
        # Empty case
        self.assertTrue(self.ts_empty.is_empty())
        # Not empty case
        self.assertFalse(self.ts_univ_1.is_empty())

    def test_extend(self) -> None:
        # Testing cases with validate=True
        # Univariate case
        self.ts_univ_extend.extend(self.ts_date_transform_univ)
        self.assertEqual(self.ts_univ_extend, self.ts_date_transform_concat_univ)
        # Multivariate case
        self.ts_multi_extend.extend(self.ts_date_transform_multi)
        self.assertEqual(self.ts_multi_extend, self.ts_date_transform_concat_multi)
        # Univariate and multivariate case
        self.ts_multi_extend_2.extend(self.ts_date_transform_univ)
        self.assertEqual(self.ts_multi_extend_2, self.ts_date_transform_concat_mixed)
        # Empty case
        self.ts_univ_default_names.extend(self.ts_empty)
        self.assertEqual(self.ts_univ_default_names, self.ts_univ_default_names_2)
        # Catching errors
        with self.assertRaises(ValueError):
            self.ts_univ_extend_err.extend(self.ts_date_transform_double_univ)
            # Multivariate case
            self.ts_multi_extend_err.extend(self.ts_date_transform_double_multi)
            # Univariate and multivariate case
            self.ts_multi_extend_err_2.extend(self.ts_date_transform_double_univ)
            # Empty case
            self.ts_empty_extend_err.extend(self.ts_empty)
        # Testing cases with validate=False
        # Univariate case
        self.ts_univ_extend_2.extend(self.ts_date_transform_double_univ, validate=False)
        self.assertEqual(self.ts_univ_extend_2, self.ts_date_transform_concat_double_univ)
        # Multivariate case
        self.ts_multi_extend_3.extend(self.ts_date_transform_double_multi, validate=False)
        self.assertEqual(self.ts_multi_extend_3, self.ts_date_transform_concat_double_multi)
        # Univariate and multivariate case
        self.ts_multi_extend_4.extend(self.ts_date_transform_double_univ, validate=False)
        self.assertEqual(self.ts_multi_extend_4, self.ts_date_transform_concat_double_mixed)
        # Empty case
        self.ts_empty_extend.extend(self.ts_empty, validate=False)
        self.assertEqual(self.ts_empty_extend, self.ts_empty)

    def test_get_item(self) -> None:
        # Univariate test case
        self.assertEqual(self.ts_date_transform_concat_univ[: len(self.ts_univ_1)], self.ts_univ_1)
        # Multivariate test case
        self.assertEqual(self.ts_date_transform_concat_multi[: len(self.ts_multi_1)], self.ts_multi_1)
        # Multivariate test case where we select a specific column
        for col in self.ts_date_transform_concat_multi.value.columns:
            ts_univ = TimeSeriesData(
                time=self.ts_date_transform_concat_multi.time,
                value=self.ts_date_transform_concat_multi.value[col],
                time_col_name=self.ts_date_transform_concat_multi.time_col_name,
            )
            self.assertEqual(self.ts_date_transform_concat_multi[col], ts_univ)
        # Multivariate test case where we select multiple columns
        self.assertEqual(
            self.ts_date_transform_concat_multi[MULTIVAR_VALUE_DF_COLS],
            self.ts_date_transform_concat_multi,
        )
        # Full/Empty cases
        self.assertEqual(self.ts_univ_1[:], self.ts_univ_1)
        self.assertEqual(
            self.ts_univ_1[0:0],
            TimeSeriesData(
                time=pd.Series(name=TIME_COL_NAME),
                value=pd.Series(name=VALUE_COL_NAME),
                time_col_name=TIME_COL_NAME,
            ),
        )


class TimeSeriesDataMiscTest(TimeSeriesBaseTest):
    def setUp(self) -> None:
        super(TimeSeriesDataMiscTest, self).setUp()
        # Creating TimeSeriesData objects
        # Univariate TimeSeriesData initialized from a pd.DataFrame
        self.ts_univ = TimeSeriesData(df=self.AIR_DF, time_col_name=TIME_COL_NAME)
        # Multivariate TimeSeriesData initialized from a pd.DataFrame
        self.ts_multi = TimeSeriesData(df=self.MULTIVAR_AIR_DF, time_col_name=TIME_COL_NAME)

    def test_is_univariate(self) -> None:
        # Univariate case
        self.assertTrue(self.ts_univ.is_univariate())
        # Multivariate case
        self.assertFalse(self.ts_multi.is_univariate())

    def test_time_to_index(self) -> None:
        # Univariate case
        assert_index_equal(self.ts_univ.time_to_index(), self.AIR_TIME_DATETIME_INDEX)
        # Multivariate case
        assert_index_equal(self.ts_multi.time_to_index(), self.AIR_TIME_DATETIME_INDEX)

    def test_repr(self) -> None:
        # Univariate case
        self.assertEqual(self.ts_univ.__repr__(), self.AIR_DF_DATETIME.__repr__())
        # Multivariate case
        self.assertEqual(self.ts_multi.__repr__(), self.MULTIVAR_AIR_DF_DATETIME.__repr__())

    def test_repr_html(self) -> None:
        # Univariate case
        self.assertEqual(self.ts_univ._repr_html_(), self.AIR_DF_DATETIME._repr_html_())
        # Multivariate case
        self.assertEqual(self.ts_multi._repr_html_(), self.MULTIVAR_AIR_DF_DATETIME._repr_html_())


class TSIteratorTest(TestCase):
    def test_ts_iterator_univariate_next(self) -> None:
        df = pd.DataFrame(
            [["2020-03-01", 100], ["2020-03-02", 120], ["2020-03-03", 130]],
            columns=["time", "y"],
        )
        kats_data = TimeSeriesData(df=df)
        kats_iterator = TSIterator(kats_data)
        val = next(kats_iterator)
        assert_series_equal(val.time, pd.Series([pd.Timestamp("2020-03-01")]), check_names=False)
        assert_series_equal(cast(pd.Series, val.value), pd.Series([100]), check_names=False)
        val = next(kats_iterator)
        assert_series_equal(val.time, pd.Series([pd.Timestamp("2020-03-02")]), check_names=False)
        assert_series_equal(cast(pd.Series, val.value), pd.Series([120]), check_names=False)
        val = next(kats_iterator)
        assert_series_equal(val.time, pd.Series([pd.Timestamp("2020-03-03")]), check_names=False)
        assert_series_equal(cast(pd.Series, val.value), pd.Series([130]), check_names=False)

    def test_ts_iterator_multivariate_next(self) -> None:
        df = pd.DataFrame(
            [
                ["2020-03-01", 100, 200],
                ["2020-03-02", 120, 220],
                ["2020-03-03", 130, 230],
            ],
            columns=["time", "y1", "y2"],
        )
        kats_data = TimeSeriesData(df=df)
        kats_iterator = TSIterator(kats_data)
        val = next(kats_iterator)
        assert_series_equal(val.time, pd.Series([pd.Timestamp("2020-03-01")]), check_names=False)
        assert_frame_equal(
            cast(pd.DataFrame, val.value),
            pd.DataFrame([[100, 200]], columns=["y1", "y2"]),
        )
        val = next(kats_iterator)
        assert_series_equal(val.time, pd.Series([pd.Timestamp("2020-03-02")]), check_names=False)
        assert_frame_equal(
            cast(pd.DataFrame, val.value),
            pd.DataFrame([[120, 220]], columns=["y1", "y2"]),
        )
        val = next(kats_iterator)
        assert_series_equal(val.time, pd.Series([pd.Timestamp("2020-03-03")]), check_names=False)
        assert_frame_equal(
            cast(pd.DataFrame, val.value),
            pd.DataFrame([[130, 230]], columns=["y1", "y2"]),
        )

    def test_ts_iterator_comprehension(self) -> None:
        kats_data = TimeSeriesData(
            time=pd.to_datetime(np.array([1596225347, 1596225348, 1596225349]), unit="s", utc=True),
            value=pd.Series(np.array([1, 2, 4])),
        )
        kats_iterator = TSIterator(kats_data)
        kats_list = list(kats_iterator)
        val = kats_list[0]
        assert_series_equal(
            val.time,
            pd.Series([pd.Timestamp("2020-07-31 19:55:47+0000", tz="UTC")]),
            check_names=False,
        )
        assert_series_equal(cast(pd.Series, val.value), pd.Series([1]), check_names=False)
        val = kats_list[1]
        assert_series_equal(
            val.time,
            pd.Series([pd.Timestamp("2020-07-31 19:55:48+0000", tz="UTC")]),
            check_names=False,
        )
        assert_series_equal(cast(pd.Series, val.value), pd.Series([2]), check_names=False)
        val = kats_list[2]
        assert_series_equal(
            val.time,
            pd.Series([pd.Timestamp("2020-07-31 19:55:49+0000", tz="UTC")]),
            check_names=False,
        )
        assert_series_equal(cast(pd.Series, val.value), pd.Series([4]), check_names=False)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any
from typing import Dict
from typing import Optional
from typing import Union
from typing import cast

import numpy as np
import pandas as pd

# Map months to seasons.
# Equivalent to ((index.month + 1) % 12) // 3 but 15x faster.
# 0: winter, 1: spring, 2: summer, 3: fall
_SEASON_MAP = {
    0: 0,
    1: 0,
    2: 1,
    3: 1,
    4: 1,
    5: 1,
    6: 2,
    7: 2,
    8: 2,
    9: 3,
    10: 3,
    11: 3,
    12: 0,
}

# Map hours to daytime periods.
# 0: sleep, 1: breakfast, 2: morning work/lunch, 3: afternoon, 4: dinnner, 5: evening
_DAYTIME_MAP = {
    0: 0,
    1: 0,
    2: 0,
    3: 0,
    4: 0,
    5: 1,
    6: 1,
    7: 1,
    8: 2,
    9: 2,
    10: 2,
    11: 2,
    12: 2,
    13: 3,
    14: 3,
    15: 3,
    16: 3,
    17: 3,
    18: 4,
    19: 4,
    20: 5,
    21: 5,
    22: 5,
    23: 0,
}


def _map(index: Union[np.ndarray, pd.Int64Index], map: Dict[int, int]) -> np.ndarray:
    """Map values to other values efficiently.

    Args:
        index: the integer index.
    Returns:
        The mapped values.
    """
    # About 4x faster than .to_series().apply(lambda).
    # About 20x faster than to_series().replace().
    values = index.to_numpy()
    result = values.copy()
    for k, v in map.items():
        result[values == k] = v
    return result


def date_features(s: pd.Series, result: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    """Compute date features for each row of a time series.

    Args:
        s: univariate time series values indexed by date.
        result: the result DataFrame to put the features into. If None, create
            one using the same index and values as s.

    Returns:
        The result with date features added.
    """
    if result is None:
        result = pd.DataFrame(s, copy=False)
    index = cast(pd.DatetimeIndex, s.index)

    result["year"] = index.year
    result["month"] = index.month
    result["day"] = index.day
    result["dayofweek"] = index.dayofweek
    result["dayofyear"] = index.dayofyear
    result["quarter"] = index.quarter
    result["season"] = _map(index.month, _SEASON_MAP)
    result["weekofyear"] = index.weekofyear
    try:
        # Work around numpy Deprecation Warning about parsing timezones
        # by converting to UTC and removing the tz info.
        dates = index.tz_convert(None).to_numpy()
    except TypeError:
        # No timezone.
        dates = index.to_numpy()
    first_of_month = pd.to_datetime(dates.astype("datetime64[M]"))
    week_of_month = np.ceil((first_of_month.dayofweek + index.day) / 7.0)
    result["weekofmonth"] = week_of_month.astype(int)
    # result["is_holiday"] = ?
    # result["holiday_types"] = ?
    result["is_weekend"] = index.dayofweek >= 5
    result["is_leap_year"] = index.is_leap_year
    result["is_leap_day"] = (index.month == 2) & (index.day == 29)
    result["is_month_end"] = index.is_month_end
    result["is_quarter_end"] = index.is_month_end & (index.month % 4 == 3)

    return result


def time_features(s: pd.Series, result: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    """Compute time features for each row of a time series.

    Args:
        s: univariate time series values indexed by time.
        result: the result DataFrame to put the features into. If None, create
            one using the same index and values as s.

    Returns:
        The result with time features added.
    """
    if result is None:
        result = pd.DataFrame(s, copy=False)
    index = cast(pd.DatetimeIndex, s.index)

    result["hour"] = index.hour
    result["minute"] = index.minute
    result["second"] = index.second
    result["milliseconds"] = index.microsecond / 1000
    result["quarterhour"] = index.minute // 15 + 1
    result["hourofweek"] = index.dayofweek * 24 + index.hour
    result["daytime"] = _map(index.hour, _DAYTIME_MAP)

    # No vectorized methods for these:
    dst = np.zeros(len(result))
    utcoffset = np.zeros(len(result))
    for i, dt in enumerate(index):
        try:
            dst[i] = dt.dst().total_seconds()
            utcoffset[i] = dt.utcoffset().total_seconds()
        except AttributeError:
            # No timezone
            dst[i] = np.nan
            utcoffset[i] = np.nan
    result["dst"] = dst
    result["utcoffset"] = utcoffset
    return result


def datetime_features(s: pd.Series, result: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    """Compute date and time features for each row of a time series.

    Args:
        s: univariate time series values indexed by datetime.
        result: the result DataFrame to put the features into. If None, create
            one using the same index and values as s.

    Returns:
        The result with date and time features added.
    """
    result = date_features(s, result)
    return time_features(s, result)


def timestamp_datetime_features(t: pd.Timestamp) -> Dict[str, Any]:
    """Compute date/time features for a single timestamp.

    Convenience method; equivalent to datetime_features() on a singleton series.

    Args:
        t: the timestamp

    Returns:
        A dictionary of features.
    """
    df = datetime_features(pd.Series([0], index=[t], name="_"))
    result = df.to_dict(orient="records")[0]
    del result["_"]
    return result


def circle_encode(ts: pd.DataFrame, features: Dict[str, int], modulo: bool = False) -> pd.DataFrame:
    """Circularly encode features.

    Args:
        ts: the multivariate time series data.
        features: dictionary of features to encode and the periodicity of those
                  features.
        modulo: if True, first reduce the features modulo the periodicity.

    Returns:
        The circularly encoded features. Each feature f generates two columns,
        f_cos and f_sin.
    """
    result = pd.DataFrame(index=ts.index)
    pi2 = np.pi * 2.0
    if modulo:
        for feat, period in features.items():
            f = (ts[feat] % period) * (pi2 / period)
            result[f"{feat}_cos"] = np.cos(f)
            result[f"{feat}_sin"] = np.sin(f)
    else:
        for feat, period in features.items():
            f = ts[feat] * (pi2 / period)
            result[f"{feat}_cos"] = np.cos(f)
            result[f"{feat}_sin"] = np.sin(f)
    return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Ensemble models with weighted average individual models

Assume we have k base models, after we make forecasts with each individual
model, we learn the weights for each individual model based on corresponding
back testing results, i.e., model with better performance should have higher
weight.
"""
import logging
import sys
from multiprocessing import Pool
from multiprocessing import cpu_count
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union
from typing import cast

import numpy as np
import pandas as pd

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.ensemble import ensemble
from kats.models.ensemble.ensemble import BASE_MODELS
from kats.models.ensemble.ensemble import EnsembleParams
from kats.models.model import Model
from kats.utils.backtesters import BackTesterSimple


class WeightedAvgEnsemble(ensemble.BaseEnsemble):
    """Weighted average ensemble model class

    Attributes:
        data: the input time series data as in :class:`kats.consts.TimeSeriesData`
        params: the model parameter class in Kats
    """

    freq: Optional[str] = None
    errors: Optional[Dict[str, Any]] = None
    weights: Optional[Dict[str, Any]] = None
    fcst_weighted: Optional[Union[pd.DataFrame, pd.Series]] = None
    fcst_dates: Optional[pd.DatetimeIndex] = None
    dates: Optional[pd.DatetimeIndex] = None
    fcst_df: Optional[pd.DataFrame] = None

    def __init__(self, data: TimeSeriesData, params: EnsembleParams) -> None:
        self.data = data
        self.params = params
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)

    def _backtester_single(
        self,
        params: Params,
        # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
        model_class: Type[Model],
        train_percentage: int = 80,
        test_percentage: int = 20,
        err_method: str = "mape",
    ) -> float:
        """Private method to run all backtesting process

        Args:
            params: Kats model parameters
            model_class: Untyped. Defines type of model
            train_percentage: float. Percentage of data used for training
            test_percentage: float. Percentage of data used for testing
            error_method: list of strings indicating which errors to calculate
                we currently support "mape", "smape", "mae", "mase", "mse", "rmse"

        Returns:
            float, the backtesting error
        """

        bt = BackTesterSimple(
            [err_method],
            self.data,
            params,
            train_percentage,
            test_percentage,
            model_class,
        )
        bt.run_backtest()
        return bt.get_error_value(err_method)

    def _backtester_all(self, err_method: str = "mape") -> Dict[str, Any]:
        """Private method to run all backtesting process

        Args:
            error_method: list of strings indicating which errors to calculate
                we currently support "mape", "smape", "mae", "mase", "mse", "rmse"

        Returns:
            Dict of errors from each model
        """

        num_process = min(len(BASE_MODELS.keys()), (cpu_count() - 1) // 2)
        if num_process < 1:
            num_process = 1
        pool = Pool(processes=(num_process), maxtasksperchild=1000)
        backtesters = {}
        for model in self.params.models:
            backtesters[model.model_name] = pool.apply_async(
                self._backtester_single,
                args=(model.model_params, BASE_MODELS[model.model_name.lower()]),
                kwds={"err_method": err_method},
            )
        pool.close()
        pool.join()
        self.errors = {model: res.get() for model, res in backtesters.items()}
        original_weights = {model: 1 / (err + sys.float_info.epsilon) for model, err in self.errors.items()}
        self.weights = {model: err / sum(original_weights.values()) for model, err in original_weights.items()}
        return self.weights

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, **kwargs) -> pd.DataFrame:
        """Predict method of weighted average ensemble model

        Args:
            steps: the length of forecasting horizon

        Returns:
            forecasting results as in pd.DataFrame
        """
        # keep these in kwargs to pass to _predict_all.
        self.freq = freq = kwargs.get("freq", "D")
        err_method = kwargs.get("err_method", "mape")
        # calculate the weights
        self._backtester_all(err_method=err_method)

        # fit model with all available time series
        pred_dict = self._predict_all(steps, **kwargs)

        fcst_all = pd.concat(
            # pyre-fixme[16]: `Model` has no attribute `fcst`.
            [x.fcst.reset_index(drop=True) for x in pred_dict.values()],
            axis=1,
            copy=False,
        )
        fcst_all.columns = cast(List[str], pred_dict.keys())
        weights = self.weights
        assert weights is not None
        self.fcst_weighted = fcst_all.dot(np.array(list(weights.values())))

        # create future dates
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=freq)
        dates = dates[dates != last_date]
        self.fcst_dates = dates.to_pydatetime()
        self.dates = dates[dates != last_date]
        self.fcst_df = fcst_df = pd.DataFrame({"time": self.dates, "fcst": self.fcst_weighted}, copy=False)

        logging.debug("Return forecast data: {fcst_df}")
        return fcst_df

    def __str__(self) -> str:
        """Get default parameter search space for the weighted average ensemble model

        Args:
            None

        Returns:
            Model name as a string
        """
        return "Weighted Average Ensemble"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file implements the Dynamic Time Warping (DTW) ChangePoint detector
algorithm as a DetectorModel, to provide a common interface.
"""

from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Type

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesChangePoint
from kats.consts import TimeSeriesData
from kats.detectors.detector import Detector

# Put code into Kats class


class DTWCPDChangePoint(TimeSeriesChangePoint):
    """Changepoint detected by the Dynamic Time Warping (DTW) detector model.

    This gives information about the type of detector, and the name of the time
    series.

    Attributes:

        start_time: Start time of the change.
        end_time: End time of the change.
        confidence: The confidence of the change point.
        ts_name: string, name of the time series for which the detector is
            is being run.
    """

    def __init__(
        self,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        confidence: float,
        ts_name: Optional[str] = None,
    ) -> None:
        super().__init__(start_time, end_time, confidence)
        self._detector_type = DTWCPDDetector
        self._ts_name = ts_name

    @property
    def detector_type(self) -> Type[Detector]:
        return self._detector_type

    @property
    def ts_name(self) -> Optional[str]:
        return self._ts_name

    def __repr__(self) -> str:
        return (
            f"DTWCPDChangePoint(start_time: {self._start_time}, end_time: "
            f"{self._end_time}, confidence: {self._confidence}, ts_name: "
            f"{self._ts_name}, detector_type={self._detector_type})"
        )


class DTWTimeSeriesTooSmallException(Exception):
    pass


@dataclass
class DTWSubsequenceMatch(object):
    matching_ts_name: str
    matching_ts_index: int
    matched_ts_name: str
    matched_ts_index: int
    distance: float


class DTWCPDDetector(Detector):
    """Dynamic Time Warping (DTW) ChangePoint detector.

    This algorithm detects anomalous time series in a dataset based on the DTW distance
    (see https://en.wikipedia.org/wiki/Dynamic_time_warping for details) between all the
    subsequences of length N in the dataset.

    Given an multivariate time series, this class performs changepoint detection, i.e. it tells
    us when the time series shows a change.

    The basic idea is to raise an alert when we see a new subsequence (via sliding window), which
    is substantially different from those seen in other time series historically. Difference here
    is measured by Dynamic Time Warping distance.

    Attrbutes:
        data: TimeSeriesData, data on which we will run the BOCPD algorithm.
        sliding_window_size: int, the time steps in the sliding window / subsequences to compare.
        mad_threshold:
        min_value:
    """

    def __init__(
        self,
        data: TimeSeriesData,
        sliding_window_size: int,
        skip_size: int,  # Step increment when sliding window moves
        mad_threshold: float = 1,  # Controls sensitivity.
        min_value: float = 1e-9,  # Controls when a value is considered to be zero
        match_against_same_time_series: bool = False,  # Whether to allow historical matches in the same time series.
    ) -> None:

        self.data = data
        self.sliding_window_size = sliding_window_size
        self.MIN_TS_VALUE = min_value
        self.MAD_CUTOFF = mad_threshold
        self.MIN_VALUE = min_value
        self.outliers: "List[DTWCPDChangePoint]" = []
        self.skip_size = skip_size
        self.match_against_same_time_series = match_against_same_time_series

        if self.sliding_window_size > len(self.data.value):
            raise DTWTimeSeriesTooSmallException(
                "The sliding window length is greater than the size of the time series."
            )

        if self.sliding_window_size > 100:
            logging.warning(
                "The model may take a very long time to run with such a large"
                f" sliding_window_size of {self.sliding_window_size}. While the model"
                " does apply fast pruning using lower bounds - it is still designed"
                " for shorter sliding_window_size"
                " (e.g., 15-50 steps), due to the underlying O(sliding_window_size^2)"
                " running time of Dynamic Time Warping. You may wish to"
                " downsample/resample the time series first."
            )

        if str(type(self.data.value)) == "<class 'pandas.core.series.Series'>":
            raise DTWTimeSeriesTooSmallException("The input time series is not multivariate")

    # pyre-fixme[14]: `detector` overrides method defined in `Detector` inconsistently.
    def detector(self) -> Sequence[DTWCPDChangePoint]:
        """Overwrite the detector method for DTWCPDDetector.

        Returns:
           A list of time series change points and matching ts information
        """

        logging.info("Starting detection...")
        # We first downsample the time series and only consider non-zero
        # sub-sequences to reduce the computational complexity of the problem.
        all_non_flats_ts = self._get_all_nonzero_subsequences()

        # If all time series are constant zero - then we are done.
        if len(all_non_flats_ts) <= 0:
            return self.outliers

        # If there is only one non-zero time series
        # then return that timeseries with the index of the first non-zero element.
        if len(all_non_flats_ts) == 1:  # There is only one time series
            ts_name = list(all_non_flats_ts.keys())[0]
            min_index = min(all_non_flats_ts[ts_name].keys())
            self.outliers.append(
                DTWCPDChangePoint(
                    start_time=self.data.time[min_index],
                    end_time=self.data.time[min_index + self.sliding_window_size - 1],
                    confidence=1e9,  # maximum confidence since its the only thing
                    ts_name=ts_name,
                )
            )
            return self.outliers  # We are done.

        # Otherwise we naively compare subsequences in each time series via an
        # efficient DTW calc to look for the closest match.
        max_min_dists = self._calculate_distances_for_all_subsequences(all_non_flats_ts)

        logging.info("Calculating anomalies")
        self.outliers = self._find_subsequences_anomalies(max_min_dists)

        return self.outliers

    def _get_nonzero_subsequences(self, ts: List[float]) -> Dict[int, List[float]]:
        win = self.sliding_window_size
        increment = self.skip_size

        return {
            ind: ts[ind : ind + win]
            for ind in range(0, len(ts) - win + 1, increment)
            if (max(ts[ind : ind + win]) > self.MIN_VALUE)
        }

    def _get_all_nonzero_subsequences(self) -> Dict[str, Dict[int, List[float]]]:
        all_non_flats_ts = {}
        df_ts = self.data.value
        c_subs = 0  # Total size of subsequences
        # TODO: vectorize + speed-up the below.
        for ix, ts_name in enumerate(df_ts.columns):
            ts = df_ts[ts_name].to_list()
            subs = self._get_nonzero_subsequences(ts)
            if len(subs) > 0:
                all_non_flats_ts[ts_name] = subs
                c_subs += len(subs)
            preprocess_log_unit = 1000
            if ix % preprocess_log_unit == 0:
                logging.info(f"Preprocessed {ix+1} / {len(df_ts.columns)}")
        logging.info(
            f"non-zero subsequences of length {self.sliding_window_size}"
            f" with minimum value of {self.MIN_TS_VALUE}: {c_subs} subsequences "
            f" on {len(all_non_flats_ts)} time series"
        )
        logging.info("Preprocessed time series.")
        return all_non_flats_ts

    @staticmethod
    def DTWDistance(s1: List[float], s2: List[float], w: int) -> float:
        """Compute DTW distance between the two lists

        Args:
            s1: first list of numbers of interest
            s2: second list of numbers of interest
            w: len(s2) // 3

        returns:
            DTW distance (a float) between the two lists
        """
        DTW = {}
        w = max(w, abs(len(s1) - len(s2)))

        for i in range(-1, len(s1)):
            for j in range(-1, len(s2)):
                DTW[(i, j)] = float("inf")
        DTW[(-1, -1)] = 0

        for i in range(len(s1)):
            for j in range(max(0, i - w), min(len(s2), i + w)):
                dist = (s1[i] - s2[j]) ** 2
                DTW[(i, j)] = dist + min(DTW[(i - 1, j)], DTW[(i, j - 1)], DTW[(i - 1, j - 1)])
        return np.sqrt(float(DTW[len(s1) - 1, len(s2) - 1]))

    @staticmethod
    def LB_Keogh(s1: List[float], s2: List[float], w: int) -> float:
        """
        Computes a lower bound on the DTW distance between two time series.

        Method according to Keogh - see # http://alexminnaar.com/2014/04/16/Time-Series-Classification-and-Clustering-with-Python.html .

        Args:
            s1: first list of numbers of interest
            s2: second list of numbers of interest
            w: a number, the lower bound on the DTW distance

        returns:
            A lower bound (float) for the DTW distance between the two given inputs.
            The lower bound is computed based on <https://www.cs.ucr.edu/~eamonn/LB_Keogh.htm>
        """
        # TODO: vectorize
        LB_sum = 0.0
        for ind, i in enumerate(s1):

            # Rolling min/max
            lower_bound = min(s2[(ind - w if ind - w >= 0 else 0) : (ind + w)])
            upper_bound = max(s2[(ind - w if ind - w >= 0 else 0) : (ind + w)])

            if i > upper_bound:
                LB_sum = LB_sum + (i - upper_bound) ** 2
            elif i < lower_bound:
                LB_sum = LB_sum + (i - lower_bound) ** 2

        return np.sqrt(LB_sum)

    def _calculate_distances_for_all_subsequences(
        self, ts2list_of_non_zero_subsequences: Dict[str, Dict[int, List[float]]]
    ) -> Dict[str, DTWSubsequenceMatch]:
        max_min_dists = defaultdict()
        for c, (ts_name_a, subsequences_a) in enumerate(ts2list_of_non_zero_subsequences.items()):
            m = []
            for inda, subsequence_a in subsequences_a.items():
                distances = []
                bsf = np.inf  # TODO: rename to upper/lower bound name
                for (
                    ts_name_b,
                    subsequences_b,
                ) in ts2list_of_non_zero_subsequences.items():
                    if (ts_name_a != ts_name_b) or self.match_against_same_time_series:
                        for indb, subsequence_b in subsequences_b.items():
                            # If it's the same time series (and we allow matching - above)
                            # Then they subsequences better not overlap - or we will
                            # obviously match.
                            if (ts_name_a == ts_name_b) and (
                                # They overlap
                                (indb <= inda <= indb + self.sliding_window_size - 1)
                                or (indb <= inda + self.sliding_window_size - 1 <= indb + self.sliding_window_size - 1)
                            ):
                                continue

                            if (
                                self.LB_Keogh(
                                    subsequence_a,
                                    subsequence_b,
                                    int(len(subsequence_b) / 3),
                                )
                                < bsf
                            ):
                                dtw_distance = self.DTWDistance(
                                    subsequence_a,
                                    subsequence_b,
                                    int(len(subsequence_b) / 3),
                                )
                                if dtw_distance < bsf:
                                    bsf = dtw_distance
                            else:
                                dtw_distance = np.inf

                            distances.append(
                                DTWSubsequenceMatch(
                                    matching_ts_name=ts_name_a,
                                    matching_ts_index=inda,
                                    matched_ts_name=ts_name_b,
                                    matched_ts_index=indb,
                                    distance=dtw_distance,
                                )
                            )
                if len(distances) > 0:
                    # finding the closest sub to query_1
                    m.append(min(distances, key=lambda x: x.distance))
            # finding the sub with largest distance
            max_min_dists[ts_name_a] = max(m, key=lambda x: x.distance)

            if c % 50 == 0:
                logging.info("%d out of %d processed ..." % (c, len(ts2list_of_non_zero_subsequences)))

        max_min_dists = dict(max_min_dists)
        return max_min_dists

    def _find_subsequences_anomalies(self, max_min_dists: Dict[str, DTWSubsequenceMatch]) -> List[DTWCPDChangePoint]:
        """
        Given a set of subsequences with at least one non-zero value,
        detect the ones that are significantly dissimilar from others.
        The dissimilarity is measured by DTW distance.
        """
        # TODO: vectorize
        outliers = []
        highest_dists = [t.distance for _, t in max_min_dists.items()]
        med = np.percentile(highest_dists, 50)
        dev_from_med = [np.abs(med - v) for v in highest_dists]
        mad = np.mean(dev_from_med)
        for ts_name, t in max_min_dists.items():
            score = np.abs(med - t.distance) / mad
            if score > self.MAD_CUTOFF:
                outliers.append(
                    DTWCPDChangePoint(
                        start_time=self.data.time[t.matching_ts_index],
                        end_time=self.data.time[t.matching_ts_index + self.sliding_window_size - 1],
                        confidence=score,
                        ts_name=ts_name,
                    )
                )
        return outliers



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# from typing import List, Optional, Union
from unittest import TestCase

from parameterized.parameterized import parameterized

from kats.metrics import metadata
from kats.metrics.metadata import MetricMetadata
from kats.metrics.metadata import inspect


class MetricsMetadataTest(TestCase):

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("continuous_rank_probability_score", metadata._ERROR_RATE_METRIC),
            ("cross_entropy", metadata._NONNEGATIVE_ERROR_METRIC),
            ("frequency_exceed_relative_threshold", metadata._ERROR_RATE_METRIC),
            ("linear_error_in_probability_space", metadata._ERROR_RATE_METRIC),
            ("max_error", metadata._UNBOUNDED_ERROR_METRIC),
            ("mean_absolute_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("mean_absolute_scaled_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("mean_error", metadata._UNBOUNDED_ERROR_METRIC),
            ("mean_interval_score", metadata._SCORE_MIN_0),
            ("mean_percentage_error", metadata._UNBOUNDED_ERROR_METRIC),
            ("mean_squared_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("mean_squared_log_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("median_absolute_error", metadata._UNBOUNDED_ERROR_METRIC),
            ("mean_absolute_percentage_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("median_absolute_percentage_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("r2_score", metadata._SCORE_MAX_1),
            ("root_mean_squared_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("root_mean_squared_log_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("root_mean_squared_percentage_error", metadata._NONNEGATIVE_ERROR_METRIC),
            ("symmetric_bias", metadata._UNBOUNDED_METRIC),
            ("symmetric_mean_absolute_percentage_error", metadata._ERROR_RATE_METRIC),
            ("tracking_signal", metadata._UNBOUNDED_ERROR_METRIC),
        ]
    )
    def test_inspect(self, metric: str, expected: MetricMetadata) -> None:
        result = inspect(metric)
        self.assertEqual(expected, result)

    def test_inspect_unknown(self) -> None:
        with self.assertRaises(ValueError):
            _ = inspect("mango")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""The Empirical Confidence (Prediction) Interval

This is an empirical way to estimate the prediction interval for any forecasting models
The high level idea is to estimate the empirical error distributions from a specific
forecasting model, and use linear regression model to fit the standard error (S.E.) with
the time horizon, under the assumption that longer horizon has larger S.E.
"""

import logging
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.utils.backtesters import BackTesterRollingWindow

FigSize = Tuple[int, int]


class EmpConfidenceInt:
    """ "class for empirical confidence interval

    The steps are listed as follows:
    1. Run K-fold CV for a given model and data,
       each fold contains h (horizon) time ahead
    2. For each horizon, calculate the Std of K error terms (S.E)
    3. Fit linear model: S.E. ~ Horizon
    4. Estimate the S.E. for each horizon for the true future
    5. Lower/Upper = Point Estimate -/+ Z_Score * S.E.

    Attributes:
        error_method: list of strings indicating which errors to calculate
            we currently support "mape", "smape", "mae", "mase", "mse", "rmse"
        data: the time series data in `TimeSeriesData` format
        params: the Kats model parameter object
        train_percentage: percentage of data used for training
        test_percentage: percentage of data used for testing
        sliding_steps: number of rolling steps to take (# of folds)
        model_class: the Kats model class
        multi: flag to use multiprocessing, the default is True
        confidence_level: the confidence level for the prediction interval
    """

    freq: Optional[str] = None
    dates: Optional[pd.DatetimeIndex] = None
    predicted: Optional[np.ndarray] = None
    coefs: Optional[np.ndarray] = None
    df: Optional[pd.DataFrame] = None
    SE: Optional[pd.DataFrame] = None

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        train_percentage: float,
        test_percentage: float,
        sliding_steps: int,
        # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
        model_class: Type,
        multi: bool = True,
        confidence_level: float = 0.8,
    ) -> None:
        self.error_methods = error_methods
        self.data = data
        self.params = params
        logging.info("Initializing train/test percentages")
        if train_percentage <= 0:
            logging.error("Non positive training percentage")
            raise ValueError("Invalid training percentage")
        elif train_percentage > 100:
            logging.error("Too large training percentage")
            raise ValueError("Invalid training percentage")
        self.train_percentage = train_percentage
        if test_percentage <= 0:
            logging.error("Non positive test percentage")
            raise ValueError("Invalid test percentage")
        elif test_percentage > 100:
            logging.error("Too large test percentage")
            raise ValueError("Invalid test percentage")
        self.test_percentage = test_percentage
        if sliding_steps < 0:
            logging.error("Non positive sliding steps")
            raise ValueError("Invalid sliding steps")
        self.sliding_steps = sliding_steps
        self.model_class = model_class
        self.multi = multi
        self.confidence_level = confidence_level
        logging.debug(
            "Initialized Empirical CI calculation with parameters. "
            "error_methods:{error_methods},"
            "data:{data},"
            "params:{params},"
            "train_percentage:{train_percentage},"
            "test_percentage:{test_percentage},"
            "sliding_steps:{sliding_steps},"
            "model_class:{model_class},"
            "multi:{multi},"
            "confidence_level:{confidence_level}".format(
                error_methods=error_methods,
                data=data,
                params=params,
                train_percentage=train_percentage,
                test_percentage=test_percentage,
                sliding_steps=sliding_steps,
                model_class=model_class,
                multi=multi,
                confidence_level=confidence_level,
            )
        )

    def run_cv(self) -> None:
        """Running the cross validation process

        run cv with given model and data
        get errors for each horizon as follows
            | horizon | error |
            |    1    |  12.3 |
            ...
        then calculate the std for each horizon
            | horizon |  std  |
            |    1    |  1.2  |
        """

        logging.info("Creating backtester object.")
        backtester = BackTesterRollingWindow(
            self.error_methods,
            self.data,
            self.params,
            self.train_percentage,
            self.test_percentage,
            self.sliding_steps,
            self.model_class,
            self.multi,
        )

        logging.info("Run backtesting.")
        backtester.run_backtest()
        self.SE = pd.DataFrame(backtester.raw_errors, copy=False).transpose().std(axis=1)

    def get_lr(self) -> None:
        """Fit linear regression model

        Fit linear regression model for
        std ~ horizon
        return the fitted model
        """
        y = self.SE
        if y is None:
            raise ValueError("Must call run_cv() before get_lr().")
        X = pd.DataFrame(
            {
                "horizon": np.arange(1, len(y) + 1),
                "const": np.ones(len(y), dtype=int),
            },
            copy=False,
        )
        logging.info("Fit the OLS model and get the coefs.")
        self.coefs = np.linalg.lstsq(X, y)[0]

    def get_eci(self, steps: int, freq: str = "D") -> pd.DataFrame:
        """Get empirical prediction interval

        Args:
            steps: the length of forecasting horizon.
            freq: the frequency.

        Returns:
            The dataframe of forecasted values with prediction intervals
        """

        self.freq = freq
        # get dates
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)
        self.dates = dates[dates != last_date]  # Return correct number of periods

        self.run_cv()
        self.get_lr()

        # run model with all data
        m = self.model_class(self.data, self.params)
        m.fit()
        self.predicted = predicted = m.predict(steps, freq=self.freq)

        # get margin of error
        horizons = np.arange(1, steps + 1)
        coefs = self.coefs
        assert coefs is not None  # set by get_lr above
        me = stats.norm.ppf(self.confidence_level) * (horizons * coefs[0] + coefs[1])

        self.df = df = pd.DataFrame(
            {
                "time": predicted.time,
                "fcst": predicted.fcst,
                "fcst_lower": predicted.fcst - me,
                "fcst_upper": predicted.fcst + me,
            },
            copy=False,
        )
        return df

    def diagnose(
        self,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[FigSize] = None,
        linecolor: str = "b",
        linelabel: str = "Fitted",
        secolor: str = "r",
        selabel: str = "Empirical S.E.",
        legend: bool = True,
    ) -> plt.Axes:
        """Diagnose the linear model fit for SE.

        Plot the OLS fit: SE ~ Horizon

        Args:
            ax: the axes to use. If None, create one.
            figsize: if creating Axes, the figsize to use. If None, use (10, 6).
            linecolor: the line color to use.
            linelabel: the legend label to use for the line plot.
            secolor: the SE color to use.
            selabel: the legend label to use for the SE plot.
            legend: if True, add the legend.

        Returns:
            The matplotlib Axes.
        """
        se = self.SE
        coefs = self.coefs
        if coefs is None:
            raise ValueError("Must call get_lr() before diagnose().")
        assert se is not None

        if ax is None:
            if figsize is None:
                figsize = (10, 6)
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()
        x = np.arange(1, len(se) + 1)
        ax.scatter(x, se, label=selabel, color=secolor)
        predictions = coefs[0] * np.array(x) + coefs[1]
        ax.plot(x, predictions, linecolor, label=linelabel)
        if legend:
            ax.legend(loc=2)
        ax.set_xlabel("Horizon")
        ax.set_ylabel("SE")
        ax.set_title("Diagnosis plot for SE ~ Horizon")
        fig.tight_layout()
        return ax

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[FigSize] = None,
        linecolor: str = "k",
        fcstcolor: str = "#4267B2",
        intervalcolor: str = "#4267B2",
        intervalalpha: float = 0.2,
        modelcolor: str = "r",
        modelalpha: float = 0.2,
        grid: bool = True,
    ) -> plt.Axes:
        """Make plot for model fitting with new uncertainty intervals.

        Args:
            ax: the axes to use. If None, create one.
            figsize: if creating Axes, the figsize to use. If None, use (10, 6).
            linecolor: the line color to use.
            fcstcolor: the line color to use for the forecast.
            intervalcolor: the fill color to use for the forecast interval.
            intervalalpha: the alpha to use for the forecast interval.
            modelcolor: the fill color to use for the model prediction interval.
            modelalpha: the alpha to use for the model prediction interval.
            grid: if True, plot gridlnes.

        Returns:
            The matplotlib Axes.
        """
        predicted = self.predicted
        if predicted is None:
            raise ValueError("Must call get_eci() before plot().")
        dates = self.dates
        df = self.df
        assert dates is not None and df is not None

        if ax is None:
            if figsize is None:
                figsize = (10, 6)
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()

        ax.plot(pd.to_datetime(self.data.time), self.data.value, linecolor)
        fcst_dates = dates.to_pydatetime()
        ax.plot(fcst_dates, df.fcst, ls="-", c=fcstcolor)

        ax.fill_between(
            fcst_dates,
            df.fcst_lower,
            df.fcst_upper,
            color=intervalcolor,
            alpha=intervalalpha,
        )

        # if there are default CI from the model, plot as well
        if predicted.shape[1] == 4:
            ax.fill_between(
                fcst_dates,
                predicted.fcst_lower,
                predicted.fcst_upper,
                color=modelcolor,
                alpha=modelalpha,
            )

        if grid:
            ax.grid(True, which="major", c="gray", ls="-", lw=1, alpha=0.2)
        ax.set_xlabel(xlabel="time")
        ax.set_ylabel(ylabel="y")
        fig.tight_layout()
        return ax



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

from kats.compat.pandas import assert_frame_equal
from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.models.sarima import SARIMAModel
from kats.models.sarima import SARIMAParams
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_SARIMA_PARAM_1_MODEL_1
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_15_SARIMA_PARAM_1_MODEL_1_INCL_HIST,
)
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_SARIMA_PARAM_1_MODEL_2
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_SARIMA_PARAM_2_MODEL_1
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_SARIMA_PARAM_2_MODEL_2
from kats.tests.models.test_models_dummy_data import AIR_FCST_30_SARIMA_PARAM_1_MODEL_1
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_30_SARIMA_PARAM_1_MODEL_1_INCL_HIST,
)
from kats.tests.models.test_models_dummy_data import AIR_FCST_30_SARIMA_PARAM_1_MODEL_2
from kats.tests.models.test_models_dummy_data import AIR_FCST_30_SARIMA_PARAM_2_MODEL_1
from kats.tests.models.test_models_dummy_data import AIR_FCST_30_SARIMA_PARAM_2_MODEL_2
from kats.tests.models.test_models_dummy_data import (
    EXOG_FCST_15_SARIMA_PARAM_EXOG_MODEL_1,
)

AIR_TS: pd.DataFrame = load_air_passengers()
MULTI_DF: pd.DataFrame = load_data("multivariate_anomaly_simulated_data.csv")
STEPS_1 = 15
STEPS_2 = 30
TEST_DATA: Dict[str, Dict[str, Any]] = {
    "monthly": {
        "ts": AIR_TS,
        "invalid_ts": TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv")),
        "freq": "MS",
        "p1": SARIMAParams(
            p=1,
            d=1,
            q=1,
        ),
        "p2": SARIMAParams(
            p=2,
            d=1,
            q=1,
            trend="ct",
            seasonal_order=(1, 0, 1, 12),
            enforce_invertibility=False,
            enforce_stationarity=False,
        ),
        "invalid_p": SARIMAParams(
            p=-1,
            d=1,
            q=1,
        ),
        "m1": {
            "cov_type": None,
            "method": "lbfgs",
            "maxiter": 50,
            "full_output": False,
            "optim_score": None,
            "optim_complex_step": False,
            "optim_hessian": None,
        },
        "m2": {
            "cov_type": None,
            "method": "newton",
            "maxiter": 1,
            "full_output": False,
            "optim_score": "harvey",
            "optim_complex_step": True,
            "optim_hessian": "opg",
        },
        "invalid_m1": {
            "start_params": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            "cov_type": "None",
            "method": "lbfgs",
            "maxiter": 50,
            "full_output": False,
            "optim_score": None,
            "optim_complex_step": False,
            "optim_hessian": None,
        },
        "invalid_m2": {
            "cov_type": None,
            "method": "invalid",
            "maxiter": 50,
            "full_output": False,
            "optim_score": None,
            "optim_complex_step": False,
            "optim_hessian": None,
        },
        "steps_1": STEPS_1,
        "steps_2": STEPS_2,
        "no_incl_hist": False,
        "incl_hist": True,
        "truth_p1_m1_15": AIR_FCST_15_SARIMA_PARAM_1_MODEL_1,
        "truth_p2_m1_15": AIR_FCST_15_SARIMA_PARAM_2_MODEL_1,
        "truth_p1_m2_15": AIR_FCST_15_SARIMA_PARAM_1_MODEL_2,
        "truth_p2_m2_15": AIR_FCST_15_SARIMA_PARAM_2_MODEL_2,
        "truth_p1_m1_30": AIR_FCST_30_SARIMA_PARAM_1_MODEL_1,
        "truth_p2_m1_30": AIR_FCST_30_SARIMA_PARAM_2_MODEL_1,
        "truth_p1_m2_30": AIR_FCST_30_SARIMA_PARAM_1_MODEL_2,
        "truth_p2_m2_30": AIR_FCST_30_SARIMA_PARAM_2_MODEL_2,
        "truth_p1_m1_15_incl_hist": AIR_FCST_15_SARIMA_PARAM_1_MODEL_1_INCL_HIST,
        "truth_p1_m1_30_incl_hist": AIR_FCST_30_SARIMA_PARAM_1_MODEL_1_INCL_HIST,
    },
}


class SARIMAModelTest(TestCase):
    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "monthly_p1_m1",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["m1"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p1_m1_15"],
                TEST_DATA["monthly"]["truth_p1_m1_30"],
                TEST_DATA["monthly"]["no_incl_hist"],
                TEST_DATA["monthly"]["freq"],
            ],
            # TODO: Figure out why results aren't deterministic: T103684646
            # [
            #     "monthly_p2_m1",
            #     TEST_DATA["monthly"]["ts"],
            #     TEST_DATA["monthly"]["p2"],
            #     TEST_DATA["monthly"]["m1"],
            #     TEST_DATA["monthly"]["steps_1"],
            #     TEST_DATA["monthly"]["steps_2"],
            #     TEST_DATA["monthly"]["truth_p2_m1_15"],
            #     TEST_DATA["monthly"]["truth_p2_m1_30"],
            #     TEST_DATA["monthly"]["no_incl_hist"],
            #     TEST_DATA["monthly"]["freq"],
            # ],
            [
                "monthly_p1_m2",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["m2"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p1_m2_15"],
                TEST_DATA["monthly"]["truth_p1_m2_30"],
                TEST_DATA["monthly"]["no_incl_hist"],
                TEST_DATA["monthly"]["freq"],
            ],
            [
                "monthly_p2_m2",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p2"],
                TEST_DATA["monthly"]["m2"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p2_m2_15"],
                TEST_DATA["monthly"]["truth_p2_m2_30"],
                TEST_DATA["monthly"]["no_incl_hist"],
                TEST_DATA["monthly"]["freq"],
            ],
            [
                "monthly_p1_m1_incl_hist",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["m1"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p1_m1_15_incl_hist"],
                TEST_DATA["monthly"]["truth_p1_m1_30_incl_hist"],
                TEST_DATA["monthly"]["incl_hist"],
                TEST_DATA["monthly"]["freq"],
            ],
        ]
    )
    def test_fcst(
        self,
        name: str,
        ts: TimeSeriesData,
        params: SARIMAParams,
        model_params: Dict[str, Optional[Union[str, int, bool]]],
        steps_1: int,
        steps_2: int,
        truth_1: pd.DataFrame,
        truth_2: pd.DataFrame,
        include_history: bool = False,
        freq: str = "MS",
    ) -> None:
        m = SARIMAModel(data=ts, params=params)
        # pyre-fixme[6]: Incompatible parameter type...
        m.fit(**model_params)
        res_1 = m.predict(steps=steps_1, include_history=include_history, freq=freq).reset_index(
            drop=True,
        )
        res_2 = m.predict(steps=steps_2, include_history=include_history, freq=freq).reset_index(
            drop=True,
        )
        assert_frame_equal(truth_1, res_1, rtol=0.01)
        assert_frame_equal(truth_2, res_2, rtol=0.01)

    def test_exog(self) -> None:
        # Prep data
        steps = STEPS_1
        df = MULTI_DF
        ts_original = TimeSeriesData(df)
        endog = df["0"][:-steps]
        time = ts_original.time_to_index()[:-steps]
        exog = df["1"][:-steps].values
        fcst_exog = df["1"][-steps:].values  # exog to be used for predictions
        ts = TimeSeriesData(value=endog, time=time)
        params = SARIMAParams(
            p=2,
            d=1,
            q=1,
            trend="ct",
            seasonal_order=(1, 0, 1, 12),
            enforce_invertibility=False,
            enforce_stationarity=False,
            exog=exog,
        )

        # Fit/Predict
        m = SARIMAModel(ts, params)
        m.fit(**TEST_DATA["monthly"]["m1"])
        res = m.predict(steps=steps, exog=fcst_exog, freq="D").reset_index(
            drop=True,
        )

        # Compare against truth
        assert_frame_equal(EXOG_FCST_15_SARIMA_PARAM_EXOG_MODEL_1, res, rtol=0.01)

        # Should raise a ValueError if exogenous variables aren't used to predict
        self.assertRaises(ValueError, m.predict, steps, "D")

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "invalid_p",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["invalid_p"],
                TEST_DATA["monthly"]["m1"],
            ],
            [
                "invalid_m1",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["invalid_m1"],
            ],
            [
                "invalid_m2",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["invalid_m2"],
            ],
            [
                "invalid_ts",
                TEST_DATA["monthly"]["invalid_ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["m1"],
            ],
        ]
    )
    def test_invalid_params(
        self,
        name: str,
        ts: TimeSeriesData,
        params: SARIMAParams,
        model_params: Dict[str, Optional[Union[str, int, bool]]],
    ) -> None:
        with self.assertRaises((ValueError, np.linalg.LinAlgError, NotImplementedError)):
            m = SARIMAModel(data=ts, params=params)
            # pyre-fixme[6]: Incompatible parameter type...
            m.fit(**model_params)

    def test_exec_plot(self) -> None:
        m = SARIMAModel(data=TEST_DATA["monthly"]["ts"], params=TEST_DATA["monthly"]["p1"])
        m.fit(**TEST_DATA["monthly"]["m1"])
        _ = m.predict(steps=STEPS_1)
        m.plot()

    def test_name(self) -> None:
        m = SARIMAModel(data=TEST_DATA["monthly"]["ts"], params=TEST_DATA["monthly"]["p1"])
        self.assertEqual(m.__str__(), "SARIMA")

    def test_search_space(self) -> None:
        m = SARIMAModel(data=TEST_DATA["monthly"]["ts"], params=TEST_DATA["monthly"]["p1"])
        self.assertEqual(
            m.get_parameter_search_space(),
            [
                {
                    "name": "p",
                    "type": "choice",
                    "values": list(range(1, 6)),
                    "value_type": "int",
                    "is_ordered": True,
                },
                {
                    "name": "d",
                    "type": "choice",
                    "values": list(range(1, 3)),
                    "value_type": "int",
                    "is_ordered": True,
                },
                {
                    "name": "q",
                    "type": "choice",
                    "values": list(range(1, 6)),
                    "value_type": "int",
                    "is_ordered": True,
                },
                {
                    "name": "seasonal_order",
                    "type": "choice",
                    "values": [
                        (1, 0, 1, 7),
                        (1, 0, 2, 7),
                        (2, 0, 1, 7),
                        (2, 0, 2, 7),
                        (1, 1, 1, 7),
                        (0, 1, 1, 7),
                    ],
                    # Note: JSON representation must be 'int', 'float', 'bool' or 'str'.
                    # so we use 'str' here instead of 'Tuple'
                    # when doing HPT, we need to convert it back to tuple
                    "value_type": "str",
                },
                {
                    "name": "trend",
                    "type": "choice",
                    "values": ["n", "c", "t", "ct"],
                    "value_type": "str",
                },
            ],
        )

    # Filler test
    def test_validate_params(self) -> None:
        params = TEST_DATA["monthly"]["p1"]
        params.validate_params()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Ensemble techniques for forecasting

This implements a set of ensemble techniques including weighted averaging, median ensemble
and STL-based ensembling method. This is the parent class for all ensemble models.
"""

import logging
from multiprocessing import Pool
from multiprocessing import cpu_count
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

import pandas as pd

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models import arima
from kats.models import holtwinters
from kats.models import linear_model
from kats.models import quadratic_model
from kats.models import sarima
from kats.models.model import Model

BASE_MODELS = {
    "arima": arima.ARIMAModel,
    "holtwinters": holtwinters.HoltWintersModel,
    "sarima": sarima.SARIMAModel,
    "linear": linear_model.LinearModel,
    "quadratic": quadratic_model.QuadraticModel,
}


class BaseModelParams(Params):
    """Ensemble parameter class

    This class contains three attributes:

    Attributes:
        model: model names (str)
        model_params: model_param is defined in base models
    """

    def __init__(self, model_name: str, model_params: Params) -> None:
        super().__init__()
        self.model_name = model_name
        self.model_params = model_params
        logging.debug(
            "Initialized Base Model parameters: "
            "Model name:{model_name},"
            "model_params:{model_params}".format(model_name=model_name, model_params=model_params)
        )

    def validate_params(self) -> None:
        logging.info("Method validate_params() is not implemented.")
        pass


class EnsembleParams(Params):
    __slots__ = ["models"]

    def __init__(self, models: List[BaseModelParams]) -> None:
        self.models: List[BaseModelParams] = models


# pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
class BaseEnsemble(Model):
    """Base ensemble class

    Implement parent class for ensemble.
    """

    # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
    fitted: Optional[Dict[str, Model]] = None

    def __init__(self, data: TimeSeriesData, params: EnsembleParams) -> None:
        super().__init__(data, params)

        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)

        for m in params.models:
            if m.model_name not in BASE_MODELS:
                msg = "Model {model_name} is not supported.\
                    Only support{models}.".format(
                    model_name=m.model_name, models=BASE_MODELS.keys()
                )
                logging.error(msg)
                raise ValueError(msg)

    def fit(self) -> None:
        """Fit method for ensemble model

        This method fits each individual model for ensembling
        and create a dict of model and fitted obj, such as
        {'m1': fitted_m1_obj, 'm2': fitted_m2_obj}
        """

        num_process = min(len(BASE_MODELS.keys()), (cpu_count() - 1) // 2)
        if num_process < 1:
            num_process = 1
        pool = Pool(processes=(num_process), maxtasksperchild=1000)

        fitted_models = {}
        for m in self.params.models:
            fitted_models[m.model_name] = pool.apply_async(
                self._fit_single,
                args=(BASE_MODELS[m.model_name.lower()], m.model_params),
            )
        pool.close()
        pool.join()
        self.fitted = {model: res.get() for model, res in fitted_models.items()}

    # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
    def _fit_single(self, model_func: Type[Model], model_param: Params) -> Model:
        """Private method to fit individual model

        Args:
            model_func: the callable model function
            model_param: the Kats model parameter class

        Returns:
            None
        """

        # get the model function call
        m = model_func(params=model_param, data=self.data)
        m.fit()
        return m

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
    def _predict_all(self, steps: int, **kwargs) -> Dict[str, Model]:
        """Private method to fit all individual models

        Args:
            steps: the length of forecasting horizon

        Returns:
            None
        """
        fitted = self.fitted
        if fitted is None:
            raise ValueError("Call fit() before predict().")

        predicted = {}
        for model_name, model_fitted in fitted.items():
            predicted[model_name] = model_fitted.predict(steps, **kwargs)
        return predicted

    def __str__(self) -> str:
        """Get the class name as a string

        Args:
            None

        Returns:
            Model name as a string
        """

        return "Ensemble"
